import os
import os.path as op
from copy import deepcopy
from typing import Union

import numpy as np
import pandas as pd
import xarray as xr

from ...additive.greensurge import (
    create_triangle_mask,
    create_triangle_mask_from_points,
    get_regular_grid,
)
from ...core.operations import nautical_to_mathematical
from .._base_wrappers import BaseModelWrapper

sbatch_file_example = """#!/bin/bash
#SBATCH --ntasks=1              # Number of tasks (MPI processes)
#SBATCH --partition=geocean     # Standard output and error log
#SBATCH --nodes=1               # Number of nodes to use
#SBATCH --mem=4gb               # Memory per node in GB (see also --mem-per-cpu)
#SBATCH --time=24:00:00

source /home/grupos/geocean/faugeree/miniforge3/etc/profile.d/conda.sh
conda activate GreenSurge

case_dir=$(ls | awk "NR == $SLURM_ARRAY_TASK_ID")
launchDelft3d.sh --case-dir $case_dir

output_file="${case_dir}/dflowfmoutput/GreenSurge_GFDcase_map.nc"
output_file_compressed="${case_dir}/dflowfmoutput/GreenSurge_GFDcase_map_compressed.nc"
output_file_compressed_tmp="${case_dir}/dflowfmoutput/GreenSurge_GFDcase_map_compressed_tmp.nc"

ncap2 -s 'mesh2d_s1=float(mesh2d_s1)' -v -O "$output_file" "$output_file_compressed_tmp" && {
  ncks -4 -L 4 "$output_file_compressed_tmp" "$output_file_compressed"
  rm "$output_file_compressed_tmp"
  [[ "$SLURM_ARRAY_TASK_ID" -ne 1 ]] && rm "$output_file"
}
"""


class Delft3dModelWrapper(BaseModelWrapper):
    """
    Wrapper for the Delft3d model.

    Attributes
    ----------
    default_parameters : dict
        The default parameters type for the wrapper.
    available_launchers : dict
        The available launchers for the wrapper.
    """

    default_parameters = {}

    available_launchers = {
        "geoocean-cluster": "launchDelft3d.sh",
        "docker_serial": "docker run --rm -v .:/case_dir -w /case_dir geoocean/rocky8 dimr dimr_config.xml",
    }

    def __init__(
        self,
        templates_dir: str,
        metamodel_parameters: dict,
        fixed_parameters: dict,
        output_dir: str,
        templates_name: dict = "all",
        debug: bool = True,
    ) -> None:
        """
        Initialize the Delft3d model wrapper.
        """

        super().__init__(
            templates_dir=templates_dir,
            metamodel_parameters=metamodel_parameters,
            fixed_parameters=fixed_parameters,
            output_dir=output_dir,
            templates_name=templates_name,
            default_parameters=self.default_parameters,
        )
        self.set_logger_name(
            name=self.__class__.__name__, level="DEBUG" if debug else "INFO"
        )

        self.sbatch_file_example = sbatch_file_example

        forcing_type = self.fixed_parameters.get("forcing_type", None)
        if forcing_type == "ASCII":
            self.fixed_parameters.setdefault(
                "ExtForceFile", "GreenSurge_GFDcase_wind_ASCII.ext"
            )
        elif forcing_type == "netCDF":
            self.fixed_parameters.setdefault(
                "ExtForceFile", "GreenSurge_GFDcase_wind_netCDF.ext"
            )

    def run_case(
        self,
        case_dir: str,
        launcher: str,
        output_log_file: str = "wrapper_out.log",
        error_log_file: str = "wrapper_error.log",
        postprocess: bool = False,
    ) -> None:
        """
        Run the case based on the launcher specified.

        Parameters
        ----------
        case_dir : str
            The case directory.
        launcher : str
            The launcher to run the case.
        output_log_file : str, optional
            The name of the output log file. Default is "wrapper_out.log".
        error_log_file : str, optional
            The name of the error log file. Default is "wrapper_error.log".
        """

        # Get launcher command from the available launchers
        launcher = self.list_available_launchers().get(launcher, launcher)

        # Run the case in the case directory
        self.logger.info(f"Running case in {case_dir} with launcher={launcher}.")
        output_log_file = op.join(case_dir, output_log_file)
        error_log_file = op.join(case_dir, error_log_file)
        self._exec_bash_commands(
            str_cmd=launcher,
            out_file=output_log_file,
            err_file=error_log_file,
            cwd=case_dir,
        )
        if postprocess:
            self.postprocess_case(case_dir=case_dir)

    def monitor_cases(
        self, dia_file_name: str, value_counts: str = None
    ) -> Union[pd.DataFrame, dict]:
        """
        Monitor the cases based on the status of the .dia files.

        Parameters
        ----------
        dia_file_name : str
            The name of the .dia file to monitor.
        """

        cases_status = {}

        for case_dir in self.cases_dirs:
            case_dir_name = op.basename(case_dir)
            case_dia_file = op.join(case_dir, dia_file_name)
            if op.exists(case_dia_file):
                with open(case_dia_file, "r") as f:
                    lines = f.readlines()
                    if any("finished" in line for line in lines[-15:]):
                        cases_status[case_dir_name] = "FINISHED"
                    else:
                        cases_status[case_dir_name] = "RUNNING"
            else:
                cases_status[case_dir_name] = "NOT STARTED"

        return super().monitor_cases(
            cases_status=cases_status, value_counts=value_counts
        )


import matplotlib.pyplot as plt


def format_matrix(mat):
    return "\n".join(
        " ".join(f"{x:.1f}" if abs(x) > 0.01 else "0" for x in line) for line in mat
    )


def format_zeros(mat_shape):
    return "\n".join("0 " * mat_shape[1] for _ in range(mat_shape[0]))


def actualize_grid_info(
    path_ds_origin: str,
    ds_GFD_calc_info: xr.Dataset,
) -> None:
    """
    Actualizes the grid information in the GFD calculation info dataset
    by adding the node coordinates and triangle connectivity from the original dataset.
    Parameters
    ----------
    path_ds_origin : str
        Path to the original dataset containing the mesh2d node coordinates.
    ds_GFD_calc_info : xr.Dataset
        The dataset containing the GFD calculation information to be updated.
    Returns
    -------
    ds_GFD_calc_info : xr.Dataset
        The updated dataset with the node coordinates and triangle connectivity.
    """

    ds_ori = xr.open_dataset(path_ds_origin)

    ds_GFD_calc_info["node_computation_longitude"] = (
        ("node_cumputation_index",),
        ds_ori.mesh2d_node_x.values,
    )
    ds_GFD_calc_info["node_computation_latitude"] = (
        ("node_cumputation_index",),
        ds_ori.mesh2d_node_y.values,
    )
    ds_GFD_calc_info["triangle_computation_connectivity"] = (
        ("element_computation_index", "triangle_forcing_nodes"),
        (ds_ori.mesh2d_face_nodes.values - 1).astype("int32"),
    )

    return ds_GFD_calc_info


class GreenSurgeModelWrapper(Delft3dModelWrapper):
    """
    Wrapper for the Delft3d model for Greensurge.
    """

    def generate_grid_forcing_file_D3DFM(
        self,
        case_context: dict,
        case_dir: str,
        ds_GFD_info: xr.Dataset,
    ):
        """
        Generate the wind files for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        ds_GFD_info : xr.Dataset
            The dataset with the GFD information.
        """
        dir_steps = case_context.get("dir_steps")
        real_dirs = np.linspace(0, 360, dir_steps + 1)[:-1]
        i_tes = case_context.get("tesela")
        i_dir = case_context.get("direction")
        real_dir = real_dirs[i_dir]
        dt_forz = case_context.get("dt_forz")
        wind_magnitude = case_context.get("wind_magnitude")
        simul_time = case_context.get("simul_time")

        node_triangle = ds_GFD_info.triangle_forcing_connectivity.isel(
            element_forcing_index=i_tes
        )
        lon_teselas = ds_GFD_info.node_forcing_longitude.isel(
            node_forcing_index=node_triangle
        ).values
        lat_teselas = ds_GFD_info.node_forcing_latitude.isel(
            node_forcing_index=node_triangle
        ).values

        lon_grid = ds_GFD_info.lon_grid.values
        lat_grid = ds_GFD_info.lat_grid.values

        x_llcenter = lon_grid[0]
        y_llcenter = lat_grid[0]

        n_cols = len(lon_grid)
        n_rows = len(lat_grid)

        dx = (lon_grid[-1] - lon_grid[0]) / n_cols
        dy = (lat_grid[-1] - lat_grid[0]) / n_rows
        X0, X1, X2 = lon_teselas
        Y0, Y1, Y2 = lat_teselas

        triangle = [(X0, Y0), (X1, Y1), (X2, Y2)]
        mask = create_triangle_mask(lon_grid, lat_grid, triangle).astype(int)
        mask_int = np.flip(mask, axis=0)  # Ojo

        u = -np.cos(nautical_to_mathematical(real_dir) * np.pi / 180) * wind_magnitude
        v = -np.sin(nautical_to_mathematical(real_dir) * np.pi / 180) * wind_magnitude
        u_mat = mask_int * u
        v_mat = mask_int * v

        self.logger.info(
            f"Creating Tecelda {i_tes} direction {int(real_dir)} with u = {u} and v = {v}"
        )

        file_name_u = op.join(case_dir, "GFD_wind_file.amu")
        file_name_v = op.join(case_dir, "GFD_wind_file.amv")

        with open(file_name_u, "w+") as fu, open(file_name_v, "w+") as fv:
            fu.write(
                "### START OF HEADER\n"
                + "### This file is created by Deltares\n"
                + "### Additional commments\n"
                + "FileVersion = 1.03\n"
                + "filetype = meteo_on_equidistant_grid\n"
                + "NODATA_value = -9999.0\n"
                + f"n_cols = {n_cols}\n"
                + f"n_rows = {n_rows}\n"
                + "grid_unit = degree\n"
                + f"x_llcenter = {x_llcenter}\n"
                + f"y_llcenter = {y_llcenter}\n"
                + f"dx = {dx}\n"
                + f"dy = {dy}\n"
                + "n_quantity = 1\n"
                + "quantity1 = x_wind\n"
                + "unit1 = m s-1\n"
                + "### END OF HEADER\n"
            )
            fv.write(
                "### START OF HEADER\n"
                + "### This file is created by Deltares\n"
                + "### Additional commments\n"
                + "FileVersion = 1.03\n"
                + "filetype = meteo_on_equidistant_grid\n"
                + "NODATA_value = -9999.0\n"
                + f"n_cols = {n_cols}\n"
                + f"n_rows = {n_rows}\n"
                + "grid_unit = degree\n"
                + f"x_llcenter = {x_llcenter}\n"
                + f"y_llcenter = {y_llcenter}\n"
                + f"dx = {dx}\n"
                + f"dy = {dy}\n"
                + "n_quantity = 1\n"
                + "quantity1 = y_wind\n"
                + "unit1 = m s-1\n"
                + "### END OF HEADER\n"
            )
            for time in range(4):
                if time == 0:
                    time_real = time
                elif time == 1:
                    time_real = dt_forz
                elif time == 2:
                    time_real = dt_forz + 0.01
                elif time == 3:
                    time_real = simul_time
                fu.write(f"TIME = {time_real} hours since 2022-01-01 00:00:00 +00:00\n")
                fv.write(f"TIME = {time_real} hours since 2022-01-01 00:00:00 +00:00\n")
                if time in [0, 1]:
                    fu.write(format_matrix(u_mat) + "\n")
                    fv.write(format_matrix(v_mat) + "\n")
                else:
                    fu.write(format_zeros(u_mat.shape) + "\n")
                    fv.write(format_zeros(v_mat.shape) + "\n")

    def generate_grid_forcing_file_netCDF_D3DFM(
        self,
        case_context: dict,
        case_dir: str,
        ds_GFD_info: xr.Dataset,
    ):
        """
        Generate the wind files for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        ds_GFD_info : xr.Dataset
            The dataset with the GFD information.
        """

        triangle_index = case_context.get("tesela")
        direction_index = case_context.get("direction")
        wind_direction = ds_GFD_info.wind_directions.values[direction_index]
        wind_speed = case_context.get("wind_magnitude")

        connectivity = ds_GFD_info.triangle_forcing_connectivity
        triangle_longitude = ds_GFD_info.node_forcing_longitude.isel(
            node_forcing_index=connectivity
        ).values
        triangle_latitude = ds_GFD_info.node_forcing_latitude.isel(
            node_forcing_index=connectivity
        ).values

        longitude_points_computation = ds_GFD_info.node_computation_longitude.values
        latitude_points_computation = ds_GFD_info.node_computation_latitude.values

        x0, x1, x2 = triangle_longitude[triangle_index, :]
        y0, y1, y2 = triangle_latitude[triangle_index, :]

        triangle_vertices = [(x0, y0), (x1, y1), (x2, y2)]
        triangle_mask = create_triangle_mask_from_points(
            longitude_points_computation, latitude_points_computation, triangle_vertices
        )

        angle_rad = nautical_to_mathematical(wind_direction) * np.pi / 180
        wind_u = -np.cos(angle_rad) * wind_speed
        wind_v = -np.sin(angle_rad) * wind_speed

        windx = np.zeros((4, len(longitude_points_computation)))
        windy = np.zeros((4, len(longitude_points_computation)))

        windx[0:2, triangle_mask] = wind_u
        windy[0:2, triangle_mask] = wind_v

        ds_forcing = ds_GFD_info[
            [
                "time_forcing_index",
                "node_cumputation_index",
                "node_computation_longitude",
                "node_computation_latitude",
            ]
        ]
        ds_forcing = ds_forcing.rename(
            {
                "time_forcing_index": "time",
                "node_cumputation_index": "node",
                "node_computation_longitude": "longitude",
                "node_computation_latitude": "latitude",
            }
        )
        ds_forcing.attrs = {}
        ds_forcing["windx"] = (("time", "node"), windx)
        ds_forcing["windy"] = (("time", "node"), windy)
        ds_forcing["windx"].attrs = {
            "coordinates": "time node",
            "long_name": "Wind speed in x direction",
            "standard_name": "windx",
            "units": "m s-1",
        }
        ds_forcing["windy"].attrs = {
            "coordinates": "time node",
            "long_name": "Wind speed in y direction",
            "standard_name": "windy",
            "units": "m s-1",
        }
        ds_forcing.to_netcdf(op.join(case_dir, "forcing.nc"))

        self.logger.info(
            f"Creating triangle {triangle_index} direction {int(wind_direction)} with u = {wind_u} and v = {wind_v}"
        )

    def build_case(
        self,
        case_context: dict,
        case_dir: str,
    ) -> None:
        """
        Build the input files for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        if case_context.get("forcing_type") == "netCDF":
            self.generate_grid_forcing_file_netCDF_D3DFM(
                case_context=case_context,
                case_dir=case_dir,
                ds_GFD_info=case_context.get("ds_GFD_info"),
            )
        elif case_context.get("forcing_type") == "ASCII":
            if case_context.get("case_num") == 0:
                ds_GFD_info = case_context.get("ds_GFD_info")
                lon_grid, lat_grid = get_regular_grid(
                    node_computation_longitude=ds_GFD_info.node_computation_longitude.values,
                    node_computation_latitude=ds_GFD_info.node_computation_latitude.values,
                    node_computation_elements=ds_GFD_info.triangle_computation_connectivity.values,
                )
                self.ds_GFD_info = deepcopy(case_context.get("ds_GFD_info"))
                self.ds_GFD_info["lon_grid"] = np.flip(lon_grid)
                self.ds_GFD_info["lat_grid"] = lat_grid

            self.generate_grid_forcing_file_D3DFM(
                case_context=case_context,
                case_dir=case_dir,
                ds_GFD_info=self.ds_GFD_info,
            )
        else:
            raise ("Unknown forcing type")

    def postprocess_case(self, case_dir: str) -> None:
        """
        Postprocess the case output file.

        Parameters
        ----------
        case_dir : str
            The case directory.
        """

        output_file = op.join(case_dir, "dflowfmoutput/GreenSurge_GFDcase_map.nc")
        output_file_compressed = op.join(
            case_dir, "dflowfmoutput/GreenSurge_GFDcase_map_compressed.nc"
        )
        output_file_compressed_tmp = op.join(
            case_dir, "dflowfmoutput/GreenSurge_GFDcase_map_compressed_tmp.nc"
        )
        if case_dir == self.output_dir[0]:
            # If the case_dir is the output_dir, we do not remove the original file
            postprocess_command = f"""
                ncap2 -s 'mesh2d_s1=float(mesh2d_s1)' -v -O "{output_file}" "{output_file_compressed_tmp}"
                ncks -4 -L 4 "{output_file_compressed_tmp}" "{output_file_compressed}"
                rm "{output_file_compressed_tmp}"
            """
        else:
            postprocess_command = f"""
                ncap2 -s 'mesh2d_s1=float(mesh2d_s1)' -v -O "{output_file}" "{output_file_compressed_tmp}"
                ncks -4 -L 4 "{output_file_compressed_tmp}" "{output_file_compressed}"
                rm "{output_file_compressed_tmp}"
                rm "{output_file}"
            """

        self._exec_bash_commands(
            str_cmd=postprocess_command,
            cwd=case_dir,
        )

    def postprocess_cases(self, ds_GFD_info: xr.Dataset, parallel: bool = False):
        """
        Postprocess the cases output files.

        Parameters
        ----------
        ds_GFD_info : xr.Dataset
            The dataset with the GFD information.
        parallel : bool, optional
            Whether to run the postprocessing in parallel. Default is False.
        """

        if (
            self.monitor_cases(
                dia_file_name="dflowfmoutput/GreenSurge_GFDcase.dia",
                value_counts="percentage",
            )
            .loc["FINISHED"]
            .values
            != 100.0
        ):
            raise ValueError(
                "Not all cases are finished. Please check the status of the cases."
            )

        path_computation = op.join(
            self.cases_dirs[0], "dflowfmoutput/GreenSurge_GFDcase_map.nc"
        )
        ds_GFD_info = actualize_grid_info(path_computation, ds_GFD_info)
        dirname, basename = os.path.split(ds_GFD_info.attrs["source"])
        name, ext = os.path.splitext(basename)
        new_filepath = os.path.join(dirname, f"{name}_updated{ext}")
        ds_GFD_info.to_netcdf(new_filepath)

        # case_ext = "dflowfmoutput/GreenSurge_GFDcase_map_compressed.nc"
        case_ext = "dflowfmoutput/GreenSurge_GFDcase_map.nc"

        def preprocess(dataset):
            file_name = dataset.encoding.get("source", None)
            dir_i = int(file_name.split("_D_")[-1].split("/")[0])
            tes_i = int(file_name.split("_T_")[-1].split("_D_")[0])
            dataset = (
                dataset[["mesh2d_s1"]]
                .expand_dims(["forcing_cell"])
                .assign_coords(forcing_cell=[tes_i])
            )
            self.logger.info(
                f"Loaded {file_name} with forcing_cell={tes_i} and dir={dir_i}"
            )
            return dataset

        folder_postprocess = op.join(self.output_dir, "GreenSurge_Postprocess")
        os.makedirs(folder_postprocess, exist_ok=True)

        dir_steps = self.fixed_parameters["dir_steps"]
        for idx in range(dir_steps):
            paths = self.cases_dirs[idx::dir_steps]
            file_paths = [op.join(case_dir, case_ext) for case_dir in paths]
            DS = xr.open_mfdataset(
                file_paths,
                parallel=parallel,
                combine="by_coords",
                preprocess=preprocess,
            )
            DS.load().to_netcdf(op.join(folder_postprocess, f"GreenSurge_DB_{idx}.nc"))



import copernicusmarine

copernicusmarine.subset(
    dataset_id="cmems_mod_glo_wav_my_0.2deg_PT3H-i",
    dataset_version="202411",
    variables=[
        "VHM0",
        "VHM0_SW1",
        "VHM0_SW2",
        "VHM0_WW",
        "VMDR",
        "VMDR_SW1",
        "VMDR_SW2",
        "VMDR_WW",
        "VPED",
        "VSDX",
        "VSDY",
        "VTM01_SW1",
        "VTM01_SW2",
        "VTM01_WW",
        "VTM02",
        "VTM10",
        "VTPK",
    ],
    minimum_longitude=-10.43452696741375,
    maximum_longitude=-0.5556814090161573,
    minimum_latitude=42.03998421470398,
    maximum_latitude=46.133428506857676,
    start_datetime="1980-01-01T00:00:00",
    end_datetime="2023-04-30T21:00:00",
    coordinates_selection_method="strict-inside",
    disable_progress_bar=False,
)



from collections import OrderedDict
from datetime import datetime
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
import statsmodels.discrete.discrete_model as sm
import xarray as xr


class AutoLogisticRegression:
    """
    Auto-Regressive Logistic (ALR) model for categorical time series modeling.

    This implementation focuses on statsmodels with AIC-based covariate selection.
    The model simulates transitions between discrete states (weather types, climate
    patterns, etc.) accounting for:
    - Markov dependencies (previous states)
    - Seasonality
    - Long-term trends
    - External covariates

    Attributes
    ----------
    cluster_size : int
        Number of discrete states or clusters in the model
    model : statsmodels.MNLogit
        Fitted multinomial logistic regression model
    terms_fit : dict
        Dictionary of terms used for model fitting
    terms_fit_names : list
        Names of the terms used for model fitting
    """

    def __init__(self, cluster_size: int):
        """
        Initialize the ALR model.

        Parameters
        ----------
        cluster_size : int
            Number of discrete states/clusters in the model
        """

        self.cluster_size = cluster_size
        self.model = None
        self.terms_fit = {}
        self.terms_fit_names = []
        self.mk_order = 0
        self.cov_names = []

    def get_year_fraction(self, time_array):
        """
        Convert datetime array to fractional year representation.

        Parameters
        ----------
        time_array : array-like
            Array of datetime objects

        Returns
        -------
        np.ndarray
            Array of year fractions (0-1 for position within the year)
        """

        time_yfrac = np.zeros(len(time_array))
        for i, dt in enumerate(time_array):
            doy = dt.timetuple().tm_yday
            year_length = 366 if dt.year % 4 == 0 else 365
            time_yfrac[i] = dt.year + doy / year_length

        return time_yfrac

    def generate_terms(
        self,
        bmus: np.ndarray,
        time: np.ndarray,
        mk_order: int = 0,
        use_constant: bool = True,
        use_long_term: bool = False,
        seasonality: Tuple[bool, List[int]] = (False, []),
        covariates: Optional[xr.Dataset] = None,
        cov_seasonality: bool = False,
    ) -> Tuple[OrderedDict, List[str]]:
        """
        Generate design matrix terms for the ALR model.

        Parameters
        ----------
        bmus : np.ndarray
            Array of categorical states (1-indexed)
        time : np.ndarray
            Array of datetime objects
        mk_order : int, optional
            Markov chain order (number of previous states to include)
        use_constant : bool, optional
            Include intercept term
        use_long_term : bool, optional
            Include long-term trend
        seasonality : tuple(bool, list), optional
            Tuple of (use_seasonality, [phases]). Example: (True, [1, 2]) for annual
            and semi-annual cycles
        covariates : xr.Dataset, optional
            Dataset with covariates. Should have dimensions (time, cov_names) and variable cov_values
        cov_seasonality : bool, optional
            Include seasonal modulation of covariates

        Returns
        -------
        OrderedDict
            Dictionary of design matrix terms
        list
            List of term names
        """

        # Convert time to year fractions
        time_yfrac = self.get_year_fraction(time)

        # Initialize containers
        terms = OrderedDict()
        terms_names = []

        # Constant term (intercept)
        if use_constant:
            terms["constant"] = np.ones((bmus.size, 1))
            terms_names.append("intercept")

        # Long-term trend
        if use_long_term:
            terms["long_term"] = np.ones((bmus.size, 1))
            terms["long_term"][:, 0] = time_yfrac
            terms_names.append("long_term")

        # Seasonality terms (harmonic functions)
        if seasonality[0]:
            phases = seasonality[1]
            temp_seas = np.zeros((len(time_yfrac), 2 * len(phases)))

            for i, harmonic in enumerate(phases):
                # Add cosine and sine components for each harmonic
                temp_seas[:, 2 * i] = np.cos(harmonic * 2 * np.pi * (time_yfrac % 1))
                temp_seas[:, 2 * i + 1] = np.sin(
                    harmonic * 2 * np.pi * (time_yfrac % 1)
                )

                terms_names.append(f"cos_{harmonic}")
                terms_names.append(f"sin_{harmonic}")

            terms["seasonality"] = temp_seas

        # External covariates
        if covariates is not None:
            cov_values = covariates.cov_values.values
            cov_names = covariates.cov_names.values
            self.cov_names = cov_names

            # Normalize covariates
            cov_norm = (cov_values - cov_values.mean(axis=0)) / cov_values.std(axis=0)

            # Add each covariate as a separate term
            for i, name in enumerate(cov_names):
                terms[name] = np.reshape(cov_norm[:, i], (cov_norm.shape[0], 1))
                terms_names.append(name)

                # Add seasonal modulation of covariates if requested
                if cov_seasonality:
                    cos_term = terms[name] * np.cos(
                        2 * np.pi * (time_yfrac % 1)
                    ).reshape(-1, 1)
                    sin_term = terms[name] * np.sin(
                        2 * np.pi * (time_yfrac % 1)
                    ).reshape(-1, 1)

                    terms[f"{name}_cos"] = cos_term
                    terms[f"{name}_sin"] = sin_term

                    terms_names.append(f"{name}_cos")
                    terms_names.append(f"{name}_sin")

        # Markov terms (previous states)
        if mk_order > 0:
            # Define Helmert contrasts for categorical variables
            def helmert_coding(k):
                """Generate Helmert contrast matrix for k categories"""
                H = np.zeros((k, k - 1))
                for i in range(k - 1):
                    H[i, i] = (k - i - 1) / (k - i)
                    H[(i + 1) :, i] = -1.0 / (k - i)
                return H

            # Generate contrast matrix
            contrast_matrix = helmert_coding(self.cluster_size)

            # Add terms for each lag in the Markov order
            for i in range(mk_order):
                # Initialize terms for this lag
                Z = np.zeros((bmus.size, self.cluster_size - 1))

                # Set values using the contrast matrix
                for j in range(bmus.size - i - 1):
                    bmu_idx = int(bmus[j]) - 1  # Convert 1-indexed to 0-indexed
                    Z[j + i + 1, :] = contrast_matrix[bmu_idx, :]

                # Add to terms
                terms[f"markov_{i + 1}"] = Z

                # Add term names
                for c in range(self.cluster_size - 1):
                    terms_names.append(f"mk{i + 1}_{c + 1}")

        return terms, terms_names

    def fit(
        self,
        bmus: np.ndarray,
        time: np.ndarray,
        mk_order: int = 0,
        use_constant: bool = True,
        use_long_term: bool = False,
        seasonality: Tuple[bool, List[int]] = (False, []),
        covariates: Optional[xr.Dataset] = None,
        cov_seasonality: bool = False,
        max_iter: int = 1000,
    ) -> None:
        """
        Fit the ALR model to data.

        Parameters
        ----------
        bmus : np.ndarray
            Array of categorical states (1-indexed)
        time : np.ndarray
            Array of datetime objects
        mk_order : int, optional
            Markov chain order (number of previous states to include)
        use_constant : bool, optional
            Include intercept term
        use_long_term : bool, optional
            Include long-term trend
        seasonality : tuple(bool, list), optional
            Tuple of (use_seasonality, [phases])
        covariates : xr.Dataset, optional
            Dataset with covariates
        cov_seasonality : bool, optional
            Include seasonal modulation of covariates
        max_iter : int, optional
            Maximum iterations for model fitting
        """

        self.mk_order = mk_order

        # Generate design matrix terms
        self.terms_fit, self.terms_fit_names = self.generate_terms(
            bmus=bmus,
            time=time,
            mk_order=mk_order,
            use_constant=use_constant,
            use_long_term=use_long_term,
            seasonality=seasonality,
            covariates=covariates,
            cov_seasonality=cov_seasonality,
        )

        # Combine terms into design matrix
        X = np.concatenate(list(self.terms_fit.values()), axis=1)
        y = bmus

        # Convert to pandas for statsmodels
        X_df = pd.DataFrame(X, columns=self.terms_fit_names)
        y_df = pd.DataFrame(y, columns=["bmus"])

        print(
            f"Fitting multinomial logistic regression with {len(self.terms_fit_names)} terms..."
        )

        # Fit the model
        self.model = sm.MNLogit(y_df, X_df).fit(
            method="lbfgs", maxiter=max_iter, disp=True
        )

        print("Model fitting complete.")

    def select_covariates_by_aic(
        self,
        bmus: np.ndarray,
        time: np.ndarray,
        covariates: xr.Dataset,
        base_config: dict = None,
    ) -> Tuple[List[str], float]:
        """
        Select optimal covariates using AIC criterion.

        Parameters
        ----------
        bmus : np.ndarray
            Array of categorical states (1-indexed)
        time : np.ndarray
            Array of datetime objects
        covariates : xr.Dataset
            Dataset with all potential covariates
        base_config : dict, optional
            Base configuration for other terms (mk_order, seasonality, etc.)

        Returns
        -------
        list
            List of selected covariate names
        float
            AIC of the best model
        """

        # Default base configuration
        if base_config is None:
            base_config = {
                "mk_order": 1,
                "use_constant": True,
                "use_long_term": False,
                "seasonality": (True, [1]),
                "cov_seasonality": False,
            }

        cov_names = covariates.cov_names.values
        n_covs = len(cov_names)

        print(f"Starting covariate selection with {n_covs} potential covariates...")

        # Initialize
        best_aic = float("inf")
        best_covs = []
        all_results = []

        # Test each covariate individually first
        for i, cov in enumerate(cov_names):
            # Create subset of covariates with just this one
            cov_subset = xr.Dataset(
                data_vars={
                    "cov_values": (
                        ("time", "cov_names"),
                        covariates.cov_values.values[:, i : i + 1],
                    ),
                },
                coords={"time": covariates.time, "cov_names": [cov]},
            )

            # Fit model
            try:
                self.fit(
                    bmus=bmus,
                    time=time,
                    mk_order=base_config["mk_order"],
                    use_constant=base_config["use_constant"],
                    use_long_term=base_config["use_long_term"],
                    seasonality=base_config["seasonality"],
                    covariates=cov_subset,
                    cov_seasonality=base_config["cov_seasonality"],
                )

                # Get AIC
                aic = self.model.aic
                all_results.append((aic, [cov]))

                print(f"Covariate {cov}: AIC = {aic:.2f}")

                # Update best model if improved
                if aic < best_aic:
                    best_aic = aic
                    best_covs = [cov]

            except Exception as e:
                print(f"Error fitting model with covariate {cov}: {str(e)}")

        # Forward selection - iteratively add variables
        current_covs = best_covs.copy()
        current_aic = best_aic

        improved = True
        while improved and len(current_covs) < n_covs:
            improved = False
            best_new_aic = float("inf")
            best_new_cov = None

            # Try adding each remaining covariate
            for cov in cov_names:
                if cov not in current_covs:
                    test_covs = current_covs + [cov]

                    # Create subset of covariates
                    cov_indices = [list(cov_names).index(c) for c in test_covs]
                    cov_subset = xr.Dataset(
                        data_vars={
                            "cov_values": (
                                ("time", "cov_names"),
                                covariates.cov_values.values[:, cov_indices],
                            ),
                        },
                        coords={"time": covariates.time, "cov_names": test_covs},
                    )

                    # Fit model
                    try:
                        self.fit(
                            bmus=bmus,
                            time=time,
                            mk_order=base_config["mk_order"],
                            use_constant=base_config["use_constant"],
                            use_long_term=base_config["use_long_term"],
                            seasonality=base_config["seasonality"],
                            covariates=cov_subset,
                            cov_seasonality=base_config["cov_seasonality"],
                        )

                        # Get AIC
                        aic = self.model.aic
                        all_results.append((aic, test_covs))

                        print(f"Covariates {test_covs}: AIC = {aic:.2f}")

                        # Update best candidate
                        if aic < best_new_aic:
                            best_new_aic = aic
                            best_new_cov = cov

                    except Exception as e:
                        print(
                            f"Error fitting model with covariates {test_covs}: {str(e)}"
                        )

            # Check if adding the best new covariate improves AIC
            if best_new_aic < current_aic:
                improved = True
                current_aic = best_new_aic
                current_covs.append(best_new_cov)

                # Update global best if improved
                if current_aic < best_aic:
                    best_aic = current_aic
                    best_covs = current_covs.copy()

                print(f"Added covariate {best_new_cov}, new AIC = {current_aic:.2f}")

        # Final model with best covariates
        if best_covs:
            print(f"\nBest model has AIC = {best_aic:.2f} with covariates: {best_covs}")

            # Create final covariate subset
            cov_indices = [list(cov_names).index(c) for c in best_covs]
            best_cov_subset = xr.Dataset(
                data_vars={
                    "cov_values": (
                        ("time", "cov_names"),
                        covariates.cov_values.values[:, cov_indices],
                    ),
                },
                coords={"time": covariates.time, "cov_names": best_covs},
            )

            # Fit final model
            self.fit(
                bmus=bmus,
                time=time,
                mk_order=base_config["mk_order"],
                use_constant=base_config["use_constant"],
                use_long_term=base_config["use_long_term"],
                seasonality=base_config["seasonality"],
                covariates=best_cov_subset,
                cov_seasonality=base_config["cov_seasonality"],
            )

        return best_covs, best_aic

    def simulate(
        self,
        time_sim: np.ndarray,
        num_sims: int = 1,
        covariates_sim: Optional[xr.Dataset] = None,
    ) -> np.ndarray:
        """
        Generate simulations from the fitted ALR model.

        Parameters
        ----------
        time_sim : np.ndarray
            Array of datetime objects for simulation
        num_sims : int, optional
            Number of simulations to generate
        covariates_sim : xr.Dataset, optional
            Covariates for simulation period

        Returns
        -------
        np.ndarray
            Array of simulated states with shape (time, num_sims)
        """

        if self.model is None:
            raise ValueError("Model must be fitted before simulation")

        # Get basic parameters
        mk_order = self.mk_order

        # Initialize output array
        sim_bmus = np.zeros((len(time_sim), num_sims))

        # Initial values (random start)
        for n in range(num_sims):
            # Initialize with random values
            init_bmus = np.random.randint(1, self.cluster_size + 1, mk_order)
            sim_bmus[:mk_order, n] = init_bmus

        # Main simulation loop
        print(f"Simulating {num_sims} time series...")

        for n in range(num_sims):
            for i in range(mk_order, len(time_sim)):
                # Extract current simulation history
                hist_bmus = sim_bmus[i - mk_order : i, n]

                # Generate terms for this step
                terms_i, _ = self.generate_terms(
                    bmus=np.append(hist_bmus, 0),
                    time=time_sim[i - mk_order : i + 1],
                    mk_order=mk_order,
                    use_constant=True,  # Assuming constant is always used
                    use_long_term="long_term" in self.terms_fit,
                    seasonality=(
                        "seasonality" in self.terms_fit,
                        [1] if "seasonality" in self.terms_fit else [],
                    ),
                    covariates=covariates_sim.sel(time=time_sim[i : i + 1])
                    if covariates_sim is not None
                    else None,
                    cov_seasonality=any(
                        k.endswith("_cos") for k in self.terms_fit.keys()
                    ),
                )

                # Prepare prediction matrix
                X = np.concatenate(list(terms_i.values()), axis=1)

                # Get transition probabilities
                probs = self.model.predict(X)
                cum_probs = np.cumsum(probs[-1, :])

                # Generate random state based on probabilities
                rnd_val = np.random.rand()
                new_bmu = np.where(cum_probs > rnd_val)[0][0] + 1

                # Store result
                sim_bmus[i, n] = new_bmu

        return sim_bmus

    def get_summary(self) -> pd.DataFrame:
        """
        Get model summary statistics.

        Returns
        -------
        pd.DataFrame
            DataFrame with parameter estimates, p-values, etc.
        """

        if self.model is None:
            raise ValueError("Model must be fitted first")

        # Extract key information
        params = self.model.params.transpose()
        pvalues = self.model.pvalues.transpose()
        conf_int = self.model.conf_int()

        # Organize into a DataFrame
        summary = pd.DataFrame()

        # for cluster in range(2, self.cluster_size + 1):
        #     cluster_params = params.iloc[cluster - 1].reset_index()
        #     cluster_params.columns = ["term", "coefficient"]

        #     cluster_pvals = pvalues.iloc[cluster - 1].reset_index()
        #     cluster_pvals.columns = ["term", "p_value"]

        # # Get confidence intervals
        # ci_lower = conf_int.xs(cluster - 1, level=1)[0].reset_index()
        # ci_upper = conf_int.xs(cluster - 1, level=1)[1].reset_index()

        # ci_lower.columns = ["term", "ci_lower"]
        # ci_upper.columns = ["term", "ci_upper"]

        # # Combine all info
        # cluster_info = cluster_params.merge(cluster_pvals, on="term")
        # cluster_info = cluster_info.merge(ci_lower, on="term")
        # cluster_info = cluster_info.merge(ci_upper, on="term")

        # # Add cluster identifier
        # cluster_info["cluster"] = cluster

        # # Add significance stars
        # cluster_info["significance"] = ""
        # cluster_info.loc[cluster_info["p_value"] < 0.05, "significance"] = "*"
        # cluster_info.loc[cluster_info["p_value"] < 0.01, "significance"] = "**"
        # cluster_info.loc[cluster_info["p_value"] < 0.001, "significance"] = "***"

        # # Append to summary
        # summary = pd.concat([summary, cluster_info])

        # Add model metrics
        print("Model Information:")
        print(f"AIC: {self.model.aic:.2f}")
        print(f"BIC: {self.model.bic:.2f}")
        print(f"Log-Likelihood: {self.model.llf:.2f}")
        print(f"Pseudo R-squared: {self.model.prsquared:.4f}")

        # return summary.sort_values(["cluster", "term"]).reset_index(drop=True)


if __name__ == "__main__":
    from datetime import datetime, timedelta

    import numpy as np
    import pandas as pd
    import xarray as xr

    # Create sample data
    np.random.seed(42)

    # Create time array
    start_date = datetime(2000, 1, 1)
    days = 1000
    time_array = np.array([start_date + timedelta(days=i) for i in range(days)])

    # Generate weather states (1-6)
    bmus = np.random.randint(1, 7, size=days)

    # Add some seasonality to states
    for i in range(days):
        day_of_year = time_array[i].timetuple().tm_yday
        if day_of_year < 100:  # Winter
            bmus[i] = np.random.choice([1, 2, 3], p=[0.6, 0.3, 0.1])
        elif day_of_year > 250:  # Fall
            bmus[i] = np.random.choice([4, 5, 6], p=[0.5, 0.3, 0.2])

    # Create some covariates
    cov1 = np.sin(np.linspace(0, 10 * np.pi, days)) + np.random.normal(
        0, 0.2, days
    )  # Oscillating
    cov2 = np.linspace(-1, 1, days) + np.random.normal(0, 0.1, days)  # Trend
    cov3 = np.random.normal(0, 1, days)  # Random noise

    # Create dataset with covariates
    covariates = xr.Dataset(
        data_vars={
            "cov_values": (("time", "cov_names"), np.column_stack([cov1, cov2, cov3])),
        },
        coords={"time": time_array, "cov_names": ["oscillation", "trend", "noise"]},
    )

    # Initialize and fit the ALR model
    alr_model = AutoLogisticRegression(cluster_size=6)

    # Select best covariates using AIC
    best_covs, best_aic = alr_model.select_covariates_by_aic(
        bmus=bmus,
        time=time_array,
        covariates=covariates,
        base_config={
            "mk_order": 1,
            "use_constant": True,
            "use_long_term": True,
            "seasonality": (True, [1, 2]),
            "cov_seasonality": True,
        },
    )

    # Get model summary
    summary = alr_model.get_summary()
    print("\nModel Parameter Summary:")
    print(summary)

    # Generate future simulation
    future_days = 100
    future_time = np.array(
        [start_date + timedelta(days=days + i) for i in range(future_days)]
    )

    # Create future covariates
    future_cov1 = np.sin(
        np.linspace(10 * np.pi, 11 * np.pi, future_days)
    ) + np.random.normal(0, 0.2, future_days)
    future_cov2 = np.linspace(1, 1.2, future_days) + np.random.normal(
        0, 0.1, future_days
    )
    future_cov3 = np.random.normal(0, 1, future_days)

    # Create dataset with future covariates (using only selected covariates)
    cov_indices = [list(covariates.cov_names.values).index(c) for c in best_covs]
    future_covs = np.column_stack(
        [[future_cov1, future_cov2, future_cov3][i] for i in cov_indices]
    )

    future_covariates = xr.Dataset(
        data_vars={
            "cov_values": (("time", "cov_names"), future_covs),
        },
        coords={"time": future_time, "cov_names": best_covs},
    )

    # Generate simulations
    simulations = alr_model.simulate(
        time_sim=future_time, num_sims=5, covariates_sim=future_covariates
    )

    print("\nSimulation Results (first 10 time steps):")
    print(simulations[:10, :])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import os
import os.path as op

# pip
import xarray as xr
import numpy as np
import netCDF4 as nc4


def ReadSLP(p_db, lat1, lat2, lon1, lon2, resample, p_save = None):
    'Read data from CFS SLP database: netCDF files'
    # TODO: poder parar y retomar la extraccion

    print('Reading SLP data from files...')

    ncfiles_1 = sorted(
        [op.join(p_db,f) for f in os.listdir(p_db) \
         if f.endswith('.nc') and 'gdas' in f]
    )
    ncfiles_2 = sorted(
        [op.join(p_db,f) for f in os.listdir(p_db) \
         if f.endswith('.nc') and 'cdas1' in f]
    )
    ncfiles = ncfiles_1 + ncfiles_2

    if not ncfiles:
        print('No files for extraction')
        return None

    # get time array
    time = []
    for f in ncfiles:
        with nc4.Dataset(f,'r') as ds:
            time_var = ds.variables['time']
            dtime = nc4.num2date(time_var[:],time_var.units)
            time += list(dtime)

    # get longitude and latitude indexes
    with nc4.Dataset(ncfiles[0],'r') as ds:
        lon = ds.variables['lon'][:]
        lat = ds.variables['lat'][:]
        ix_lon1 = np.where(lon==lon1)[0][0]
        ix_lon2 = np.where(lon==lon2)[0][0]
        ix_lat1 = np.where(lat==lat1)[0][0]
        ix_lat2 = np.where(lat==lat2)[0][0]
        longitude = lon[ix_lon1:ix_lon2+resample:resample]
        latitude = lat[ix_lat1:ix_lat2+resample:resample]

    # read slp file by file
    np_SLP = np.nan * np.ones((len(time), len(latitude), len(longitude)))
    ti = 0
    for f in ncfiles:
        print(f)
        with nc4.Dataset(f,'r') as ds:
            time_len = len(ds.variables['time'])
            slp = ds.variables['PRMSL_L101'][:]
            np_SLP[ti:ti+time_len,:,:] = slp[
                :,
                ix_lat1:ix_lat2+resample:resample,
                ix_lon1:ix_lon2+resample:resample
            ]
            ti += time_len

    # mount output dataset
    xds_SLP = xr.Dataset(
        {
            'SLP': (('time','latitude','longitude'), np_SLP),
        },
        coords = {
            'time': time,
            'longitude': longitude,
            'latitude': latitude,
        },
    )

    # save
    if p_save:
        xds_SLP.to_netcdf(p_save)

    return xds_SLP





import numpy as np
import xarray as xr


def generate_grid_parameters(
    bathy_data: xr.DataArray,
    alpc: float = 0,
    xpc: float = None,
    ypc: float = None,
    xlenc: float = None,
    ylenc: float = None,
    buffer_distance: float = None,
) -> dict:
    """
    Generate grid parameters for the SWAN model based on bathymetry.

    Parameters
    ----------
    bathy_data : xr.DataArray
                Bathymetry data.
        Must have the following dimensions:
        - lon/x: longitude or x coordinate
        - lat/y: latitude or y coordinate
    alpc: float
        Computational Grid Rotation angle in degrees.
    xpc: float
        X origin.
    ypc: float
        Y origin.

    Returns
    -------
    dict
        Dictionary with grid configuration for SWAN input.
    """

    # Determine coordinate system based on coordinate names
    coord_names = list(bathy_data.coords)

    # Get coordinate variables
    if any(name in ["lon", "longitude"] for name in coord_names):
        x_coord = next(name for name in coord_names if name in ["lon", "longitude"])
        y_coord = next(name for name in coord_names if name in ["lat", "latitude"])
    else:
        x_coord = next(
            name for name in coord_names if name in ["x", "X", "cx", "easting"]
        )
        y_coord = next(
            name for name in coord_names if name in ["y", "Y", "cy", "northing"]
        )

    # Get resolution from cropped data
    grid_resolution_x = abs(
        bathy_data[x_coord][1].values - bathy_data[x_coord][0].values
    )
    grid_resolution_y = abs(
        bathy_data[y_coord][1].values - bathy_data[y_coord][0].values
    )

    if alpc != 0:
        angle_rad = np.radians(alpc)
        # Create rotation matrix
        R = np.array(
            [
                [np.cos(angle_rad), -np.sin(angle_rad)],
                [np.sin(angle_rad), np.cos(angle_rad)],
            ]
        )

        # Create unrotated rectangle corners
        dx = np.array([0, xlenc, xlenc, 0, 0])
        dy = np.array([0, 0, ylenc, ylenc, 0])
        points = np.column_stack([dx, dy])

        # Rotate points
        rotated = np.dot(points, R.T)

        # Translate to corner position
        x = rotated[:, 0] + xpc
        y = rotated[:, 1] + ypc
        corners = np.column_stack([x, y])

        x_min = np.min(x) - buffer_distance
        x_max = np.max(x) + buffer_distance
        y_min = np.min(y) - buffer_distance
        y_max = np.max(y) + buffer_distance

        print(f"Cropping bounds: x=[{x_min}, {x_max}], y=[{y_min}, {y_max}]")

        # Crop bathymetry
        cropped = bathy_data.sel(
            {
                x_coord: slice(x_min, x_max),
                y_coord: slice(y_max, y_min),
            }  # Note: slice from max to min for descending coordinates
        )

        grid_parameters = {
            "xpc": xpc,
            "ypc": ypc,
            "alpc": alpc,
            "xlenc": xlenc,
            "ylenc": ylenc,
            "mxc": int(np.round(xlenc / grid_resolution_x) - 1),
            "myc": int(np.round(ylenc / grid_resolution_y) - 1),
            "xpinp": np.nanmin(cropped[x_coord]),  # x origin from cropped data
            "ypinp": np.nanmin(cropped[y_coord]),  # y origin from cropped data
            "alpinp": 0,  # x-axis direction
            "mxinp": len(cropped[x_coord]) - 1,  # number mesh x from cropped data
            "myinp": len(cropped[y_coord]) - 1,  # number mesh y from cropped data
            "dxinp": grid_resolution_x,  # resolution from cropped data
            "dyinp": grid_resolution_y,  # resolution from cropped data
        }
        return grid_parameters, cropped, corners

    else:
        # Compute parameters from full bathymetry
        grid_parameters = {
            "xpc": float(np.nanmin(bathy_data[x_coord])),  # origin x
            "ypc": float(np.nanmin(bathy_data[y_coord])),  # origin y
            "alpc": alpc,  # x-axis direction
            "xlenc": float(
                np.nanmax(bathy_data[x_coord]) - np.nanmin(bathy_data[x_coord])
            ),  # grid length x
            "ylenc": float(
                np.nanmax(bathy_data[y_coord]) - np.nanmin(bathy_data[y_coord])
            ),  # grid length y
            "mxc": len(bathy_data[x_coord]) - 1,  # num mesh x
            "myc": len(bathy_data[y_coord]) - 1,  # num mesh y
            "xpinp": float(np.nanmin(bathy_data[x_coord])),  # origin x
            "ypinp": float(np.nanmin(bathy_data[y_coord])),  # origin y
            "alpinp": 0,  # x-axis direction
            "mxinp": len(bathy_data[x_coord]) - 1,  # num mesh x
            "myinp": len(bathy_data[y_coord]) - 1,  # num mesh y
            "dxinp": float(
                abs(bathy_data[x_coord][1].values - bathy_data[x_coord][0].values)
            ),  # resolution x
            "dyinp": float(
                abs(bathy_data[y_coord][1].values - bathy_data[y_coord][0].values)
            ),  # resolution y
        }
        return grid_parameters



"""
Project: BlueMath_tk
Sub-Module: distributions
Author: GeoOcean Research Group, Universidad de Cantabria
Repository: https://github.com/GeoOcean/BlueMath_tk.git
Status: Under development (Working)
"""

# Import essential functions/classes to be available at the package level.
from .gev import GEV
from .gpd import GPD
from .nonstat_gev import NonStatGEV

# Optionally, define the module's `__all__` variable to control what gets imported when using `from module import *`.
__all__ = ["GEV", "GPD", "NonStatGEV"]



import os.path as op
from typing import Dict

from siphon.catalog import TDSCatalog

GEOOCEAN_CLUSTER_DATA = "/lustre/geocean/DATA/"
GEOOCEAN_THREDDS_DATA = "https://geoocean.sci.unican.es/thredds/dodsC/geoceanData/"

# Default paths dictionary
PATHS = {
    "SHYTCWAVES_COEFS": op.join(
        GEOOCEAN_THREDDS_DATA,
        "GEOOCEAN/SHyTCWaves_bulk/ibtracs_coef_pmin_wmax.nc",
    ),
    "SHYTCWAVES_BULK": op.join(
        GEOOCEAN_THREDDS_DATA,
        "GEOOCEAN/SHyTCWaves_bulk/library_shytcwaves_bulk_params_int32.nc",
    ),
    "SHYTCWAVES_MDA": op.join(
        GEOOCEAN_THREDDS_DATA,
        "GEOOCEAN/SHyTCWaves_bulk/shytcwaves_mda.nc",
    ),
    "SHYTCWAVES_MDA_INDICES": op.join(
        GEOOCEAN_THREDDS_DATA,
        "GEOOCEAN/SHyTCWaves_bulk/shytcwaves_mda_indices.nc",
    ),
    "SHYTCWAVES_MDA_MASK_INDICES": op.join(
        GEOOCEAN_THREDDS_DATA,
        "GEOOCEAN/SHyTCWaves_bulk/mda_mask_indices.nc",
    ),
    "SHYTCWAVES_MDA_MASK_INDICES_LOWRES": op.join(
        GEOOCEAN_THREDDS_DATA,
        "GEOOCEAN/SHyTCWaves_bulk/mda_mask_indices_lowres.nc",
    ),
    "SHYTCWAVES_SPECTRA": op.join(
        GEOOCEAN_CLUSTER_DATA,
        "GEOOCEAN/SHyTCWaves/",
    ),
}


def update_paths(new_paths: dict) -> None:
    """
    Update the paths dictionary with new values.

    Parameters
    ----------
    new_paths : dict
        Dictionary containing new path values to update.

    Examples
    --------
    >>> update_paths({"MY_PATH": "/new/path/to/data"})
    """

    PATHS.update(new_paths)


def get_paths(verbose: bool = True) -> dict:
    """
    Get the paths dictionary.

    Parameters
    ----------
    verbose : bool, optional
        Whether to print warnings about Thredds paths. Default is True.

    Returns
    -------
    dict
        Dictionary containing the paths.
    """

    if verbose:
        for file_name, file_path in PATHS.items():
            if "thredds" in file_path:
                print(f"WARNING: {file_name} is a Thredds path.")
            elif "lustre" in file_path:
                print(f"WARNING: {file_name} is a GeoOcean cluster path.")
                print("* Data access can be requested to bluemath@unican.es")
        print(
            "You can update any path or add new paths with the update_paths function,"
            " from bluemath_tk.config.paths."
        )
        print("Example: update_paths({'SHYTCWAVES_COEFS': '/new/path/to/data'})")

    return PATHS


def get_thredds_catalog() -> TDSCatalog:
    """
    Get the Thredds catalog object.

    Returns
    -------
    TDSCatalog
        Siphon TDSCatalog object containing the catalog information.
    """

    catalog_url = (
        "https://geoocean.sci.unican.es/thredds/catalog/geoceanData/catalog.xml"
    )

    return TDSCatalog(catalog_url)


def get_catalog_folders() -> Dict[str, str]:
    """
    Get a dictionary of folder names and their links from the first level of the catalog.

    Returns
    -------
    Dict[str, str]
        Dictionary with folder names as keys and their catalog URLs as values.
    """

    catalog = get_thredds_catalog()
    folders = {}

    for name, ref in catalog.catalog_refs.items():
        folders[ref.title] = ref.href

    return folders


def print_catalog_table() -> None:
    """
    Print a formatted table of available folders in the catalog.
    """

    folders = get_catalog_folders()

    # Print header
    print("\nAvailable Folders in GeoOcean Thredds Catalog:")
    print("-" * 80)
    print(f"{'Folder Name':<20} | {'Catalog URL':<40}")
    print("-" * 80)

    # Print each folder
    for name, url in sorted(folders.items()):
        print(f"{name:<20} | {url:<40}")

    print("-" * 80)


if __name__ == "__main__":
    print_catalog_table()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from math import radians, degrees, sin, cos, asin, acos, sqrt, atan2, pi
import numpy as np
import xarray as xr

from .util.operations import GetUniqueRows


def GeoDistance(lat1, lon1, lat2, lon2):
    'Returns great circle distance between points in degrees'

    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

    a = sin((lat2-lat1)/2)**2 + cos(lat1) * cos(lat2) * sin((lon2-lon1)/2)**2;
    if a < 0: a = 0
    if a > 1: a = 1

    r = 1
    rng = r * 2 * atan2(sqrt(a), sqrt(1-a))
    rng = degrees(rng)

    return rng

def GeoAzimuth(lat1, lon1, lat2, lon2):
    'Returns geodesic azimuth between point1 and point2'

    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

    az = atan2(
        cos(lat2) * sin(lon2-lon1),
        cos(lat1) * sin(lat2) - sin(lat1) * cos(lat2) * cos(lon2-lon1)
    )
    if lat1 <= -pi/2: az = 0
    if lat2 >=  pi/2: az = 0
    if lat2 <= -pi/2: az = pi
    if lat1 >=  pi/2: az = pi

    az = az % (2*pi)
    az = degrees(az)

    return az


def Extract_Circle(xds_TCs, p_lon, p_lat, r, d_vns):
    '''
    Extracts TCs inside circle - used with NWO or Nakajo databases

    xds_TCs: tropical cyclones track database
        lon, lat, pressure variables
        storm dimension

    circle defined by:
        p_lon, p_lat  -  circle center
        r             -  circle radius (degree)

    d_vns: dictionary to set longitude, latitude, time and pressure varnames

    returns:
        xds_area: selection of xds_TCs inside circle
        xds_inside: contains TCs custom data inside circle
    '''

    # point longitude and latitude
    lonlat_p = np.array([[p_lon, p_lat]])

    # get names of vars: longitude, latitude, pressure and time
    nm_lon = d_vns['longitude']
    nm_lat = d_vns['latitude']
    nm_prs = d_vns['pressure']
    nm_tim = d_vns['time']

    # storms longitude, latitude, pressure and time (if available)
    lon = xds_TCs[nm_lon].values[:]
    lat = xds_TCs[nm_lat].values[:]
    prs = xds_TCs[nm_prs].values[:]
    time = xds_TCs[nm_tim].values[:]

    # get storms inside circle area
    n_storms = xds_TCs.storm.shape[0]
    l_storms_area = []

    # inside parameters holders
    l_prs_min_in = []   # circle minimun pressure
    l_prs_mean_in = []  # circle mean pressure
    l_vel_mean_in = []  # circle mean translation speed
    l_categ_in = []     # circle storm category
    l_date_in = []      # circle date (day)
    l_date_last = []    # last cyclone date 
    l_gamma = []        # azimuth 
    l_delta = []        # delta 

    l_ix_in = []        # historical enters the circle index
    l_ix_out = []       # historical leaves the circle index 

    for i_storm in range(n_storms):

        # fix longitude <0 data and skip "one point" tracks
        lon_storm = lon[i_storm]
        if not isinstance(lon_storm, np.ndarray): continue
        lon_storm[lon_storm<0] = lon_storm[lon_storm<0] + 360

        # stack storm longitude, latitude
        lonlat_s = np.column_stack(
            (lon_storm, lat[i_storm])
        )

        # index for removing nans
        ix_nonan = ~np.isnan(lonlat_s).any(axis=1)
        lonlat_s = lonlat_s[ix_nonan]

        # calculate geodesic distance (degree)
        geo_dist = []
        for lon_ps, lat_ps in lonlat_s:
            geo_dist.append(GeoDistance(lat_ps, lon_ps, p_lat, p_lon))
        geo_dist = np.asarray(geo_dist)
        
        # find storm inside circle and calculate parameters
        if (geo_dist < r).any():

            # storm inside circle
            ix_in = np.where(geo_dist < r)[0][:]

            # storm translation velocity
            geo_dist_ss = []
            for i_row in range(lonlat_s.shape[0]-1):
                i0_lat, i0_lon = lonlat_s[i_row][1], lonlat_s[i_row][0]
                i1_lat, i1_lon = lonlat_s[i_row+1][1], lonlat_s[i_row+1][0]
                geo_dist_ss.append(GeoDistance(i0_lat, i0_lon, i1_lat, i1_lon))
            geo_dist_ss = np.asarray(geo_dist_ss)

            # get delta time in hours (irregular data time delta)
            if isinstance(time[i_storm][0], np.datetime64):
                # round to days
                time[i_storm] = np.array(
                    [np.datetime64(xt, 'h') for xt in time[i_storm]]
                )

                delta_h = np.diff(
                    time[i_storm][~np.isnat(time[i_storm])]
                ).astype('timedelta64[h]').astype(float)

            else:
                # nakajo: time already in hours
                delta_h = np.diff(
                    time[i_storm][~np.isnan(lon[i_storm])]
                ).astype(float)

            vel = geo_dist_ss * 111.0/delta_h  # km/h

            # promediate vel 
            velpm = (vel[:-1] + vel[1:])/2
            velpm = np.append(vel[0], velpm)
            velpm = np.append(velpm, vel[-1])

            # calculate azimuth 
            lat_in_end, lon_in_end = lonlat_s[ix_in[-1]][1], lonlat_s[ix_in[-1]][0]
            lat_in_ini, lon_in_ini = lonlat_s[ix_in[0]][1], lonlat_s[ix_in[0]][0]
            gamma = GeoAzimuth(lat_in_end, lon_in_end, lat_in_ini, lon_in_ini)
            if gamma < 0.0: gamma += 360

            # calculate delta
            nd = 1000
            st = 2*np.pi/nd
            ang = np.arange(0, 2*np.pi + st, st)
            xps = r * np.cos(ang) + p_lat
            yps = r * np.sin(ang) + p_lon
            angle_radius = []
            for x, y in zip(xps, yps):
                angle_radius.append(GeoAzimuth(lat_in_end, lon_in_end, x, y))
            angle_radius = np.asarray(angle_radius)

            im = np.argmin(np.absolute(angle_radius - gamma))
            delta = GeoAzimuth(p_lat, p_lon, xps[im], yps[im]) # (-180, +180)
            if delta < 0.0: delta += 360

            # more parameters 
            prs_s_in = prs[i_storm][ix_in]  # pressure

            # nan data filter
            if np.all(np.isnan(prs_s_in)):
                continue
            no_nan = ~np.isnan(prs_s_in)

            prs_s_in = prs_s_in[no_nan]
            prs_s_min = np.min(prs_s_in)  # pressure minimun
            prs_s_mean = np.mean(prs_s_in)

            vel_s_in = velpm[ix_in][no_nan]  # velocity
            vel_s_mean = np.mean(vel_s_in) # velocity mean

            categ = GetStormCategory(prs_s_min)  # category

            dist_in = geo_dist[ix_in][no_nan]
            p_dm = np.where((dist_in==np.min(dist_in)))[0]  # closest to point

            time_s_in = time[i_storm][ix_in][no_nan]  # time
            time_closest = time_s_in[p_dm][0]  # time closest to point 

            # filter storms 
            # TODO: storms with only one track point inside radius. solve?
            if np.isnan(np.array(prs_s_in)).any() or \
               (np.array(prs_s_in) <= 860).any() or \
               gamma == 0.0:
                continue

            # store parameters
            l_storms_area.append(i_storm)
            l_prs_min_in.append(np.array(prs_s_min))
            l_prs_mean_in.append(np.array(prs_s_mean))
            l_vel_mean_in.append(np.array(vel_s_mean))
            l_categ_in.append(np.array(categ))
            l_date_in.append(time_closest)
            l_gamma.append(gamma)
            l_delta.append(delta)

            # store historical indexes inside circle 
            l_ix_in.append(ix_in[0])
            l_ix_out.append(ix_in[-1])

            # store last cyclone date too
            l_date_last.append(time[i_storm][ix_nonan][-1])

    # cut storm dataset to selection
    xds_TCs_sel = xds_TCs.isel(storm=l_storms_area)
    xds_TCs_sel = xds_TCs_sel.assign_coords(storm = np.array(l_storms_area))

    # store storms parameters 
    xds_TCs_sel_params = xr.Dataset(
        {
            'pressure_min':(('storm'), np.array(l_prs_min_in)),
            'pressure_mean':(('storm'), np.array(l_prs_mean_in)),
            'velocity_mean':(('storm'), np.array(l_vel_mean_in)),
            'gamma':(('storm'), np.array(l_gamma)),
            'delta':(('storm'), np.array(l_delta)),
            'category':(('storm'), np.array(l_categ_in)),
            'dmin_date':(('storm'), np.array(l_date_in)),
            'last_date':(('storm'), np.array(l_date_last)),
            'index_in':(('storm'), np.array(l_ix_in)),
            'index_out':(('storm'), np.array(l_ix_out)),
        },
        coords = {
            'storm':(('storm'), np.array(l_storms_area))
        },
        attrs = {
            'point_lon' : p_lon,
            'point_lat' : p_lat,
            'point_r' : r,
        }
    )

    return xds_TCs_sel, xds_TCs_sel_params

def GetStormCategory(pres_min):
    '''
    Returns storm category (int 5-0)
    '''

    pres_lims = [920, 944, 964, 979, 1000]

    if pres_min <= pres_lims[0]:
        return 5
    elif pres_min <= pres_lims[1]:
        return 4
    elif pres_min <= pres_lims[2]:
        return 3
    elif pres_min <= pres_lims[3]:
        return 2
    elif pres_min <= pres_lims[4]:
        return 1
    else:
        return 0

def SortCategoryCount(np_categ, nocat=9):
    '''
    Sort category change - count matrix
    np_categ = [[category1, category2, count], ...]
    '''

    categs = [0,1,2,3,4,5,9]

    np_categ = np_categ.astype(int)
    np_sort = np.empty((len(categs)*(len(categs)-1),3))
    rc=0
    for c1 in categs[:-1]:
        for c2 in categs:
            p_row = np.where((np_categ[:,0]==c1) & (np_categ[:,1]==c2))
            if p_row[0].size:
                np_sort[rc,:]=[c1,c2,np_categ[p_row,2]]
            else:
                np_sort[rc,:]=[c1,c2,0]

            rc+=1

    return np_sort.astype(int)

def GetCategoryChangeProbs(xds_r1, xds_r2):
    'Calculates category change probabilities from r1 to r2'

    # Get storm category inside both circles
    n_storms = len(xds_r1.storm)
    categ_r1r2 = np.empty((n_storms, 2))
    for i in range(len(xds_r1.storm)):

        # category inside R1
        storm_in_r1 = xds_r1.isel(storm=[i])
        storm_id = storm_in_r1.storm.values[0]
        storm_cat_r1 = storm_in_r1.category

        # category inside R2
        if storm_id in xds_r2.storm.values[:]:
            storm_in_r2 = xds_r2.sel(storm=[storm_id])
            storm_cat_r2 = storm_in_r2.category
        else:
            storm_cat_r2 = 9  # no category

        # store categories
        categ_r1r2[i,:] = [storm_cat_r1, storm_cat_r2]

    # count category changes and sort it
    categ_count = GetUniqueRows(categ_r1r2)
    categ_count = SortCategoryCount(categ_count)

    # calculate probability
    m_count = np.reshape(categ_count[:,2], (6,-1)).T
    m_sum = np.sum(m_count,axis=0)

    probs = m_count.astype(float)/m_sum.astype(float)
    probs_cs = np.cumsum(probs, axis=0)

    # TODO: category_change_sum ??

    # store output using xarray
    xds_categ_cp = xr.Dataset(
        {
            'category_change_count': (('category','category'), m_count[:-1,:]),
            #'category_change_sum': (('category'), m_count[-1,:]),
            'category_change_probs': (('category','category'), probs[:-1,:]),
            'category_nochange_probs': (('category'), probs[-1,:]),
            'category_change_cumsum': (('category','category'), probs_cs[:-1,:]),
        },
        coords = {
            'category': [0,1,2,3,4,5]
        }
    )

    return xds_categ_cp




import os
import re
from typing import List, Union

import numpy as np
import pandas as pd
import scipy.io as sio
import wavespectra
import xarray as xr
from wavespectra.construct import construct_partition

from .._base_wrappers import BaseModelWrapper
from .._utils_wrappers import write_array_in_file


class SwanModelWrapper(BaseModelWrapper):
    """
    Wrapper for the SWAN model.
    https://swanmodel.sourceforge.io/online_doc/swanuse/swanuse.html

    Attributes
    ----------
    default_parameters : dict
        The default parameters type for the wrapper.
    available_launchers : dict
        The available launchers for the wrapper.
    output_variables : dict
        The output variables for the wrapper.
    """

    default_parameters = {
        "Hs": {
            "type": float,
            "value": None,
            "description": "Significant wave height.",
        },
        "Tp": {
            "type": float,
            "value": None,
            "description": "Wave peak period.",
        },
        "Dir": {
            "type": float,
            "value": None,
            "description": "Wave direction.",
        },
        "Spr": {
            "type": float,
            "value": None,
            "description": "Directional spread.",
        },
        "dir_dist": {
            "type": str,
            "choices": ["CIRCLE", "SECTOR"],
            "value": "CIRCLE",
            "description": "CIRCLE indicates that the spectral directions cover the full circle. SECTOR indicates that the spectral directions cover a limited sector of the circle.",
        },
        "dir1": {
            "type": float,
            "value": None,
            "description": "Only with SECTOR option. The direction of the right-hand boundary of the sector when looking outward from the sector (in degrees).",
        },
        "dir2": {
            "type": float,
            "value": None,
            "description": "Only with SECTOR option. The direction of the left-hand boundary of the sector when looking outward from the sector (in degrees).",
        },
        "mdc": {
            "type": int,
            "value": 24,
            "description": "Spectral directional discretization.",
        },
        "flow": {
            "type": float,
            "value": 0.03,
            "description": "Low values for frequency.",
        },
        "fhigh": {
            "type": float,
            "value": 0.5,
            "description": "High value for frequency.",
        },
        "Freq_array": {
            "type": np.ndarray,
            "value": None,
            "description": "Array of frequencies for the model.",
        },
        "Dir_array": {
            "type": np.ndarray,
            "value": None,
            "description": "Array of directions for the model.",
        },
    }

    available_launchers = {
        "serial": "swan_serial.exe",
        "docker_serial": "docker run --rm -v .:/case_dir -w /case_dir geoocean/rocky8 swan_serial.exe",
        "geoocean-cluster": "launchSwan.sh",
    }

    output_variables = {
        "Depth": {
            "long_name": "Water depth at the point",
            "units": "m",
        },
        "Hsig": {
            "long_name": "Significant wave height",
            "units": "m",
        },
        "Tm02": {
            "long_name": "Mean wave period",
            "units": "s",
        },
        "Dir": {
            "long_name": "Wave direction",
            "units": "degrees",
        },
        "PkDir": {
            "long_name": "Peak wave direction",
            "units": "degrees",
        },
        "TPsmoo": {
            "long_name": "Peak wave period",
            "units": "s",
        },
        "Dspr": {
            "long_name": "Directional spread",
            "units": "degrees",
        },
    }

    def __init__(
        self,
        templates_dir: str,
        metamodel_parameters: dict,
        fixed_parameters: dict,
        output_dir: str,
        templates_name: dict = "all",
        depth_array: np.ndarray = None,
        locations: np.ndarray = None,
        debug: bool = True,
    ) -> None:
        """
        Initialize the SWAN model wrapper.
        """

        super().__init__(
            templates_dir=templates_dir,
            metamodel_parameters=metamodel_parameters,
            fixed_parameters=fixed_parameters,
            output_dir=output_dir,
            templates_name=templates_name,
            default_parameters=self.default_parameters,
        )
        self.set_logger_name(
            name=self.__class__.__name__, level="DEBUG" if debug else "INFO"
        )

        if depth_array is not None:
            self.depth_array = np.round(depth_array, 5)
        else:
            self.depth_array = None

        if locations is not None:
            self.locations = np.column_stack(locations)
        else:
            self.locations = None

    def list_available_output_variables(self) -> List[str]:
        """
        List available output variables.

        Returns
        -------
        List[str]
            The available output variables.
        """

        return list(self.output_variables.keys())

    def _convert_case_output_files_to_nc(
        self, case_num: int, output_path: str, output_vars: List[str]
    ) -> xr.Dataset:
        """
        Convert mat file to netCDF file.

        Parameters
        ----------
        case_num : int
            The case number.
        output_path : str
            The output path.
        output_vars : List[str]
            The output variables to use.

        Returns
        -------
        xr.Dataset
            The xarray Dataset.
        """

        # Read mat file
        output_dict = sio.loadmat(output_path)

        # Create Dataset
        ds_output_dict = {var: (("Yp", "Xp"), output_dict[var]) for var in output_vars}
        ds = xr.Dataset(
            ds_output_dict,
            coords={"Xp": output_dict["Xp"][0, :], "Yp": output_dict["Yp"][:, 0]},
        )

        # assign correct coordinate case_num
        ds.coords["case_num"] = case_num

        return ds

    def get_case_percentage_from_file(self, output_log_file: str) -> str:
        """
        Get the case percentage from the output log file.

        Parameters
        ----------
        output_log_file : str
            The output log file.

        Returns
        -------
        str
            The case percentage.
        """

        if not os.path.exists(output_log_file):
            return "0 %"

        progress_pattern = r"OK in\s+(\d+\.\d+)\s*%"
        with open(output_log_file, "r") as f:
            for line in reversed(f.readlines()):
                match = re.search(progress_pattern, line)
                if match:
                    if float(match.group(1)) > 99.5:
                        return "100 %"
                    return f"{match.group(1)} %"

        return "0 %"  # if no progress is found

    def monitor_cases(self, value_counts: str = None) -> Union[pd.DataFrame, dict]:
        """
        Monitor the cases based on the wrapper_out.log file.
        """

        cases_status = {}

        for case_dir in self.cases_dirs:
            output_log_file = os.path.join(case_dir, "wrapper_out.log")
            progress = self.get_case_percentage_from_file(
                output_log_file=output_log_file
            )
            cases_status[os.path.basename(case_dir)] = progress

        return super().monitor_cases(
            cases_status=cases_status, value_counts=value_counts
        )

    def postprocess_case(
        self,
        case_num: int,
        case_dir: str,
        output_vars: List[str] = ["Hsig", "Tm02", "Dir"],
    ) -> xr.Dataset:
        """
        Convert mat ouput files to netCDF file.

        Parameters
        ----------
        case_num : int
            The case number.
        case_dir : str
            The case directory.
        output_vars : list, optional
            The output variables to postprocess. Default is None.

        Returns
        -------
        xr.Dataset
            The postprocessed Dataset.
        """

        if output_vars is None:
            self.logger.info("Postprocessing all available variables.")
            output_vars = list(self.output_variables.keys())

        output_nc_path = os.path.join(case_dir, "output.nc")
        if not os.path.exists(output_nc_path):
            # Convert tab files to netCDF file
            output_path = os.path.join(case_dir, "output.mat")
            output_nc = self._convert_case_output_files_to_nc(
                case_num=case_num,
                output_path=output_path,
                output_vars=output_vars,
            )
            output_nc.to_netcdf(os.path.join(case_dir, "output.nc"))
        else:
            self.logger.info("Reading existing output.nc file.")
            output_nc = xr.open_dataset(output_nc_path)

        return output_nc

    def join_postprocessed_files(
        self, postprocessed_files: List[xr.Dataset]
    ) -> xr.Dataset:
        """
        Join postprocessed files in a single Dataset.

        Parameters
        ----------
        postprocessed_files : list
            The postprocessed files.

        Returns
        -------
        xr.Dataset
            The joined Dataset.
        """

        return xr.concat(postprocessed_files, dim="case_num")


def generate_fixed_parameters(
    grid_parameters: dict,
    freq_array: np.array,
    dir_array: np.array,
) -> dict:
    """
    Generate fixed parameters for the SWAN model based on grid parameters and frequency/direction arrays.
    Parameters
    ----------
    grid_parameters : dict
        Dictionary with grid configuration for SWAN input.
    freq_array : np.ndarray
        Array of frequencies for the SWAN model.
    dir_array : np.ndarray
        Array of directions for the SWAN model.
    Returns
    -------
    dict
        Dictionary with fixed parameters for the SWAN model.
    """

    dirs = np.sort(np.unique(dir_array)) % 360
    step = np.round(np.median(np.diff(np.sort(dirs))), 4)

    # Compute angular gaps between sorted directions (including wrap-around)
    diffs = np.diff(np.concatenate([dirs, [dirs[0] + 360]]))
    max_gap_idx = np.argmax(diffs)

    if np.isclose(diffs[max_gap_idx], step, atol=1e-2):
        dir_dist = "CIRCLE"
        dir1, dir2 = None, None
    else:
        dir_dist = "SECTOR"
        dir1 = float((dirs[(max_gap_idx + 1) % len(dirs)]) % 360)  # right-hand boundary
        dir2 = float((dirs[max_gap_idx]) % 360)  # left-hand boundary
    print("Distribución direccional:", dir_dist)
    if dir_dist == "SECTOR":
        print(f"Direcciones de {dir1}° a {dir2}°")

    return {
        "xpc": grid_parameters["xpc"],  # origin x
        "ypc": grid_parameters["ypc"],  # origin y
        "alpc": grid_parameters["alpc"],  # x-axis direction
        "xlenc": grid_parameters["xlenc"],  # grid length x
        "ylenc": grid_parameters["ylenc"],  # grid length y
        "mxc": grid_parameters["mxc"],  # num mesh x
        "myc": grid_parameters["myc"],  # num mesh y
        "xpinp": grid_parameters["xpinp"],  # origin x for input grid
        "ypinp": grid_parameters["ypinp"],  # origin y for input grid
        "alpinp": grid_parameters["alpinp"],  # x-axis direction
        "mxinp": grid_parameters["mxinp"],  # num mesh x for input grid
        "myinp": grid_parameters["myinp"],  # num mesh y for input grid
        "dxinp": grid_parameters["dxinp"],  # resolution x for input grid
        "dyinp": grid_parameters["dyinp"],  # resolution y for input grid
        "dir_dist": dir_dist,  # direction distribution type
        "dir1": dir1,  # min direction
        "dir2": dir2,  # max direction
        "freq_discretization": len(np.unique(freq_array)),  # frequency discretization
        "dir_discretization": int(
            360 / (np.unique(dir_array)[1] - np.unique(dir_array)[0])
        ),  # direction discretization
        "mdc": int(
            360 / (np.unique(dir_array)[1] - np.unique(dir_array)[0])
        ),  # number of depth cases
        "flow": float(np.min(np.unique(freq_array))),  # low frequency limit
        "fhigh": float(np.max(np.unique(freq_array))),  # high frequency limit
    }


class BinWavesWrapper(SwanModelWrapper):
    """
    Wrapper example for the BinWaves model.
    """

    def build_case(self, case_dir: str, case_context: dict) -> None:
        if self.depth_array is not None:
            write_array_in_file(self.depth_array, f"{case_dir}/depth.dat")
        if self.locations is not None:
            write_array_in_file(self.locations, f"{case_dir}/locations.loc")

        # Construct the input spectrum
        input_spectrum = construct_partition(
            freq_name="jonswap",
            freq_kwargs={
                "freq": np.geomspace(
                    case_context.get("flow", 0.035),
                    case_context.get("fhigh", 0.5),
                    case_context.get("freq_discretization", 29),
                ),
                # "freq": np.linspace(case_context.get("flow", 0.035), case_context.get("fhigh", 0.5), case_context.get("freq_discretization", 29)),
                "fp": 1.0 / case_context.get("tp"),
                "hs": case_context.get("hs"),
            },
            dir_name="cartwright",
            dir_kwargs={
                "dir": np.linspace(0, 360, case_context.get("dir_discretization", 24)),
                "dm": case_context.get("dir"),
                "dspr": case_context.get("spr"),
            },
        )
        argmax_bin = np.argmax(input_spectrum.values)
        mono_spec_array = np.zeros(input_spectrum.freq.size * input_spectrum.dir.size)
        mono_spec_array[argmax_bin] = input_spectrum.sum(dim=["freq", "dir"])
        mono_spec_array = mono_spec_array.reshape(
            input_spectrum.freq.size, input_spectrum.dir.size
        )
        mono_input_spectrum = xr.Dataset(
            {
                "efth": (["freq", "dir"], mono_spec_array),
            },
            coords={
                "freq": input_spectrum.freq,
                "dir": input_spectrum.dir,
            },
        )
        for side in ["N", "S", "E", "W"]:
            wavespectra.SpecDataset(mono_input_spectrum).to_swan(
                os.path.join(case_dir, f"input_spectra_{side}.bnd")
            )



import tempfile
import unittest

from bluemath_tk.downloaders.copernicus.copernicus_downloader import (
    CopernicusDownloader,
)


class TestCopernicusDownloader(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.downloader = CopernicusDownloader(
            product="ERA5",
            base_path_to_download=self.temp_dir,
            token=None,
            check=True,  # Just check paths to download
        )

    def test_download_data_era5(self):
        result = self.downloader.download_data_era5(
            variables=["spectra"],
            years=[f"{year:04d}" for year in range(2020, 2025)],
            months=[
                "01",
                "02",
                "03",
                "04",
                "05",
                "06",
                "07",
                "08",
                "09",
                "10",
                "11",
                "12",
            ],
            area=[43.4, 350.4, 43.6, 350.6],  # [lat_min, lon_min, lat_max, lon_max]
        )
        print(result)


if __name__ == "__main__":
    unittest.main()


# mean_wave_period_based_on_first_moment/
# wave_spectral_directional_width/
# wave_spectral_directional_width_for_swell/
# wave_spectral_directional_width_for_wind_waves/



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import os
import os.path as op
import itertools
import calendar
from datetime import datetime

# pip
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.dates as mdates
from mpl_toolkits.mplot3d import Axes3D

# teslakit
from ..util.operations import GetBestRowsCols
from ..util.time_operations import xds_reindex_daily as xr_daily
from ..util.time_operations import xds_common_dates_daily as xcd_daily
from ..util.time_operations import get_years_months_days
from ..kma import ClusterProbabilities
from .custom_colors import colors_awt
from .pcs import axplot_PC_hist, axplot_PCs_3D


# import constants
from .config import _faspect, _fsize, _fdpi

def axplot_AWT_2D(ax, var_2D, num_wts, id_wt, color_wt):
    'axes plot AWT variable (2D)'

    # plot 2D AWT
    ax.pcolormesh(
        var_2D,
        cmap='RdBu_r', shading='gouraud',
        vmin=-1.5, vmax=+1.5,
    )

    # title and axis labels/ticks
    ax.set_title(
        'WT #{0} --- {1} years'.format(id_wt, num_wts),
        {'fontsize': 14, 'fontweight':'bold'}
    )
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_ylabel('month', {'fontsize':8})
    ax.set_xlabel('lon', {'fontsize':8})

    # set WT color on axis frame
    plt.setp(ax.spines.values(), color=color_wt, linewidth=4)
    plt.setp([ax.get_xticklines(), ax.get_yticklines()], color=color_wt)

def axplot_AWT_years(ax, dates_wt, bmus_wt, color_wt, xticks_clean=False,
                     ylab=None, xlims=None):
    'axes plot AWT dates'

    # date axis locator
    yloc5 = mdates.YearLocator(5)
    yloc1 = mdates.YearLocator(1)
    yfmt = mdates.DateFormatter('%Y')

    # get years string
    ys_str = np.array([str(d).split('-')[0] for d in dates_wt])

    # use a text bottom - top cycler
    text_cycler_va = itertools.cycle(['bottom', 'top'])
    text_cycler_ha = itertools.cycle(['left', 'right'])

    # plot AWT dates and bmus
    ax.plot(
        dates_wt, bmus_wt,
        marker='+',markersize=9, linestyle='', color=color_wt,
    )
    va = 'bottom'
    for tx,ty,tt in zip(dates_wt, bmus_wt, ys_str):
        ax.text(
            tx, ty, tt,
            {'fontsize':8},
            verticalalignment = next(text_cycler_va),
            horizontalalignment = next(text_cycler_ha),
            rotation=45,
        )

    # configure axis
    ax.set_yticks([])
    ax.xaxis.set_major_locator(yloc5)
    ax.xaxis.set_minor_locator(yloc1)
    ax.xaxis.set_major_formatter(yfmt)
    ax.grid(True, which='both', axis='x', linestyle='--', color='grey')
    ax.tick_params(axis='x', which='major', labelsize=8)

    # optional parameters
    if xticks_clean:
        ax.set_xticklabels([])
    else:
        ax.set_xlabel('Year', {'fontsize':8})

    if ylab: ax.set_ylabel(ylab)

    if xlims is not None:
        ax.set_xlim(xlims[0], xlims[1])

def axplot_EOF_evolution(ax, years, EOF_evol):
    'axes plot EOFs evolution'

    # date axis locator
    yloc5 = mdates.YearLocator(5)
    yloc1 = mdates.YearLocator(1)
    yfmt = mdates.DateFormatter('%Y')

    # get years datetime
    ys_dt = np.array([datetime(y,1,1) for y in years])

    # plot EOF evolution 
    ax.plot(
        ys_dt, EOF_evol,
        linestyle='-', color='black',
    )

    # configure axis
    ax.set_xlim(ys_dt[0], ys_dt[-1])
    ax.xaxis.set_major_locator(yloc5)
    ax.xaxis.set_minor_locator(yloc1)
    ax.xaxis.set_major_formatter(yfmt)
    ax.grid(True, which='both', axis='x', linestyle='--', color='grey')
    ax.tick_params(axis='both', which='major', labelsize=8)

def axplot_EOF(ax, EOF_value, lon, ylbl, ttl):
    'axes plot EOFs evolution'

    # EOF pcolormesh 
    ax.pcolormesh(
        lon, range(12), np.transpose(EOF_value),
        cmap='RdBu_r', shading='gouraud',
        clim=2,
    )

    # axis and title
    ax.set_yticklabels(ylbl)
    ax.set_title(
        ttl,
        {'fontsize': 14, 'fontweight':'bold'}
    )
    ax.tick_params(axis='x', which='major', labelsize=8)
    ax.tick_params(axis='y', which='major', labelsize=10)


def Plot_AWT_Validation_Cluster(AWT_2D, AWT_num_wts, AWT_ID, AWT_dates,
                                AWT_bmus, AWT_PCs_fit, AWT_PCs_rnd, AWT_color,
                                show=True):


    # figure
    fig = plt.figure(figsize=(_faspect*_fsize, _fsize))

    # layout
    gs = gridspec.GridSpec(4, 4, wspace=0.10, hspace=0.15)
    ax_AWT_2D = plt.subplot(gs[:2, :2])
    ax_PCs3D_fit = plt.subplot(gs[2, 0], projection='3d')
    ax_PCs3D_rnd = plt.subplot(gs[2, 1], projection='3d')
    ax_AWT_y = plt.subplot(gs[3, :])
    ax_PC1_hst_fit = plt.subplot(gs[0, 2])
    ax_PC1_hst_rnd = plt.subplot(gs[0, 3])
    ax_PC2_hst_fit = plt.subplot(gs[1, 2])
    ax_PC2_hst_rnd = plt.subplot(gs[1, 3])
    ax_PC3_hst_fit = plt.subplot(gs[2, 2])
    ax_PC3_hst_rnd = plt.subplot(gs[2, 3])

    # plot AWT 2D
    axplot_AWT_2D(ax_AWT_2D, AWT_2D, AWT_num_wts, AWT_ID, AWT_color)

    # plot AWT years
    axplot_AWT_years(ax_AWT_y, AWT_dates, AWT_bmus, AWT_color)

    # compare PCs fit - sim with 3D plot
    axplot_PCs_3D(ax_PCs3D_fit, AWT_PCs_fit,  AWT_color, ttl='PCs fit')
    axplot_PCs_3D(ax_PCs3D_rnd, AWT_PCs_rnd,  AWT_color, ttl='PCs sim')

    # compare PC1 histograms
    axplot_PC_hist(ax_PC1_hst_fit, AWT_PCs_fit[:,0], AWT_color)
    axplot_PC_hist(ax_PC1_hst_rnd, AWT_PCs_rnd[:,0], AWT_color, ylab='PC1')

    axplot_PC_hist(ax_PC2_hst_fit, AWT_PCs_fit[:,1], AWT_color)
    axplot_PC_hist(ax_PC2_hst_rnd, AWT_PCs_rnd[:,1], AWT_color, ylab='PC2')

    axplot_PC_hist(ax_PC3_hst_fit, AWT_PCs_fit[:,2], AWT_color)
    axplot_PC_hist(ax_PC3_hst_rnd, AWT_PCs_rnd[:,2], AWT_color, ylab='PC3')

    # show
    if show: plt.show()
    return fig

def Plot_AWTs_Validation(bmus, dates, Km, n_clusters, lon, d_PCs_fit,
                         d_PCs_rnd, show=True):
    '''
    Plot Annual Weather Types Validation

    bmus, dates, Km, n_clusters, lon - from KMA_simple()
    d_PCs_fit, d_PCs_rnd - historical and simulated PCs by WT
    '''

    # get cluster colors
    cs_awt = colors_awt()

    # each cluster has a figure
    l_figs = []
    for ic in range(n_clusters):

        # get cluster data
        id_AWT = ic + 1           # cluster ID
        index = np.where(bmus==ic)[0][:]
        dates_AWT = dates[index]  # cluster dates
        bmus_AWT = bmus[index]    # cluster bmus
        var_AWT = Km[ic,:]
        var_AWT_2D = var_AWT.reshape(-1, len(lon))
        num_WTs = len(index)      # number of cluster ocurrences
        clr = cs_awt[ic]          # cluster color
        PCs_fit = d_PCs_fit['{0}'.format(id_AWT)]
        PCs_rnd = d_PCs_rnd['{0}'.format(id_AWT)]

        # plot cluster figure
        fig = Plot_AWT_Validation_Cluster(
            var_AWT_2D, num_WTs, id_AWT,
            dates_AWT, bmus_AWT,
            PCs_fit, PCs_rnd,
            clr, show=show)

        l_figs.append(fig)

    return l_figs

def Plot_AWTs(bmus, Km, n_clusters, lon, show=True):
    '''
    Plot Annual Weather Types

    bmus, Km, n_clusters, lon - from KMA_simple()
    '''

    # get number of rows and cols for gridplot 
    n_cols, n_rows = GetBestRowsCols(n_clusters)

    # get cluster colors
    cs_awt = colors_awt()

    # plot figure
    fig = plt.figure(figsize=(_faspect*_fsize, _fsize))

    gs = gridspec.GridSpec(n_rows, n_cols, wspace=0.10, hspace=0.15)
    gr, gc = 0, 0

    for ic in range(n_clusters):

        id_AWT = ic + 1           # cluster ID
        index = np.where(bmus==ic)[0][:]
        var_AWT = Km[ic,:]
        var_AWT_2D = var_AWT.reshape(-1, len(lon))
        num_WTs = len(index)
        clr = cs_awt[ic]          # cluster color

        # AWT var 2D 
        ax = plt.subplot(gs[gr, gc])
        axplot_AWT_2D(ax, var_AWT_2D, num_WTs, id_AWT, clr)

        gc += 1
        if gc >= n_cols:
            gc = 0
            gr += 1

    # show and return figure
    if show: plt.show()
    return fig

def Plot_AWTs_Dates(bmus, dates, n_clusters, show=True):
    '''
    Plot Annual Weather Types dates

    bmus, dates, n_clusters - from KMA_simple()
    '''

    # get cluster colors
    cs_awt = colors_awt()

    # plot figure
    fig, axs = plt.subplots(nrows=n_clusters, figsize=(_faspect*_fsize, _fsize))

    # each cluster has a figure
    for ic in range(n_clusters):

        id_AWT = ic + 1           # cluster ID
        index = np.where(bmus==ic)[0][:]
        dates_AWT = dates[index]  # cluster dates
        bmus_AWT = bmus[index]    # cluster bmus
        clr = cs_awt[ic]          # cluster color

        ylabel = "WT #{0}".format(id_AWT)
        xlims = [dates[0].astype('datetime64[Y]')-np.timedelta64(3, 'Y'), dates[-1].astype('datetime64[Y]')+np.timedelta64(3, 'Y')]

        xaxis_clean=True
        if ic == n_clusters-1:
            xaxis_clean=False

        # axs plot
        axplot_AWT_years(
            axs[ic], dates_AWT, bmus_AWT,
            clr, xaxis_clean, ylabel, xlims
        )

    # show and return figure
    if show: plt.show()
    return fig

def Plot_AWTs_EOFs(PCs, EOFs, variance, time, lon, n_plot, show=True):
    '''
    Plot annual EOFs for PCA_LatitudeAverage predictor

    PCs, EOFs, variance, time, lon - from PCA_LatitudeAverage()
    n_plot                         - number of EOFs plotted
    '''

    # transpose
    EOFs = np.transpose(EOFs)
    PCs = np.transpose(PCs)

    # get start and end month
    ys, ms, _ = get_years_months_days(time)

    # PCA latavg metadata
    y1 = ys[0]
    y2 = ys[-1]
    m1 = ms[0]
    m2 = ms[-1]

    # mesh data
    len_x = len(lon)

    # time data
    years = range(y1, y2+1)
    l_months = [calendar.month_name[x] for x in range(1,13)]
    ylbl = l_months[m1-1:] + l_months[:m2]

    # percentage of variance each field explains
    n_percent = variance / np.sum(variance)

    l_figs = []
    for it in range(n_plot):

        # map of the spatial field
        spatial_fields = EOFs[:,it]*np.sqrt(variance[it])

        # reshape from vector to matrix with separated months
        C = np.reshape(
            spatial_fields[:len_x*12], (12, len_x)
        ).transpose()

        # plot figure
        fig = plt.figure(figsize=(_faspect*_fsize, _fsize))

        # layout
        gs = gridspec.GridSpec(4, 4, wspace=0.10, hspace=0.2)
        ax_EOF = plt.subplot(gs[:3, :])
        ax_evol = plt.subplot(gs[3, :])

        # EOF pcolormesh
        ttl = 'EOF #{0}  ---  {1:.2f}%'.format(it+1, n_percent[it]*100)
        axplot_EOF(ax_EOF, C, lon, ylbl, ttl)

        # time series EOF evolution
        evol =  PCs[it,:]/np.sqrt(variance[it])
        axplot_EOF_evolution(ax_evol, years, evol)

        l_figs.append(fig)

        # show
        if show: plt.show()

    return l_figs




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# pip
import numpy as np
import xarray as xr

# tk
from .util.time_operations import npdt64todatetime

# TODO: refactor con waves.py/hydrographs

class Hydrograph(object):
    'Stores hydrograph data'

    def __init__(self):
        self.date_index = []
        self.dates = []

        self.indx_hydro = []
        self.numdays_hydro = []
        self.Hs_hydro = []
        self.Tp_hydro = []
        self.Dir_hydro = []

        self.TWL_max = []
        self.Hs_max = []

        self.MU = []
        self.TAU = []


def Calculate_Hydrographs(xds_BMUS, xds_WAVES):
    '''
    Calculates intradaily hydrographs

    xds_BMUS: (time) bmus
    xds_WAVES: (time) hs, tp, dir

    returns dictionary of Hydrograph objects for each WT
    and list of xarray.Datasets containing MU and TAU
    '''

    # solve intradaily bins
    bmus = xds_BMUS.bmus.values[:]
    time_KMA = xds_BMUS.time.values[:]

    d_bins = {}
    l_mutau_xdsets = []
    for i_wt in sorted(set(bmus)):

        # find WT indexes at KMA bmus
        indx = np.where((bmus == i_wt))[0]
        date_index = time_KMA[indx]

        # find hydrograms longer than 1 day
        diff_time = np.array([
            (b-a).astype('timedelta64[D]')/np.timedelta64(1,'D') \
            for a,b in zip(date_index, date_index[1:])])
        sep_hydro = np.where((diff_time > 1.0))[0]

        if len(sep_hydro)==0:
            bin_k = 'WT {0:02d}'.format(i_wt)
            d_bins[bin_k] = None
            print('{0} empty'.format(bin_k))
            continue

        hydro_indx = []
        hydro_indx.append(indx[0:sep_hydro[0]+1])
        for m in range(len(sep_hydro)-2):
            hydro_indx.append(
                indx[sep_hydro[m]+1:sep_hydro[m+1]+1]
            )
        hydro_indx.append([indx[sep_hydro[len(sep_hydro)-1]-1]])
        num_days = [len(x) for x in hydro_indx]

        # initialize some output lists
        Hs_hydro = []
        Tp_hydro = []
        Dir_hydro = []
        TWL_max = []
        Hs_max = []
        MU = []
        TAU = []

        # work with storms <= 4 days
        ndays_storms = 4
        for h, nd in zip(hydro_indx, num_days):

            # initialize loop output
            hs_W = []
            tp_W = []
            dir_W = []
            twl_max_v = 0
            twl_max_t = 0
            hs_max_v = 0
            mu_v = 0

            if nd <= ndays_storms:

                # start and end dates for hydrograph
                p1 = npdt64todatetime(time_KMA[h[0]])
                p2 = npdt64todatetime(time_KMA[h[-1]]).replace(hour=23)

                # get waves conditions for hydrograph
                xds_W = xds_WAVES.sel(time = slice(p1, p2))
                hs_W = xds_W.Hs.values[:]
                tp_W = xds_W.Tp.values[:]
                dir_W = xds_W.Dir.values[:]

                # calculate TWL max and normalize
                twl_temp = 0.1*(hs_W**0.5)*tp_W

                dt = 1.0/len(twl_temp)
                t_norm = np.arange(dt,1+dt,dt)

                i_twlmax = np.where(twl_temp == np.amax(twl_temp))[0]
                twl_max_v =  twl_temp[i_twlmax][0]
                twl_max_t =  t_norm[i_twlmax][0]

                # calculate MU
                mu_v = np.trapz(np.divide(twl_temp, twl_max_v), t_norm)

                # calculate max Hs and Tp
                hs_max_v = hs_W[np.where(hs_W == np.amax(hs_W))[0]][0]
                tp_max_v = tp_W[np.where(tp_W == np.amax(tp_W))[0]][0]

            # store values
            Hs_hydro.append(hs_W)
            Tp_hydro.append(tp_W)
            Dir_hydro.append(dir_W)
            TWL_max.append(twl_max_v)
            Hs_max.append(hs_max_v)
            MU.append(mu_v)
            TAU.append(twl_max_t)

        # bin key and hydrograph storage
        bin_k = 'bin{0:02d}'.format(i_wt)
        bin_hy = Hydrograph()

        bin_hy.date_index = indx
        bin_hy.dates = date_index
        bin_hy.indx_hydro = hydro_indx
        bin_hy.numdays_hydro = num_days
        bin_hy.Hs_hydro = Hs_hydro
        bin_hy.Tp_hydro = Tp_hydro
        bin_hy.Dir_hydro = Dir_hydro
        bin_hy.TWL_max = TWL_max
        bin_hy.Hs_max = Hs_max
        bin_hy.MU = MU
        bin_hy.TAU = TAU

        # store at dictionary
        d_bins[bin_k] = bin_hy
        #print('{0} calculated'.format(bin_k))

        # store mu, tau xarray.Dataset
        xds_hg_wt = xr.Dataset(
            {
                'MU':(('time',), np.array(MU)),
                'TAU':(('time',), np.array(TAU)),
                'hs_max':(('time',), np.array(Hs_max)),
                'twl_max':(('time',), np.array(TWL_max)),
            },
            attrs={'WT':i_wt+1}
        )
        l_mutau_xdsets.append(xds_hg_wt)

    return d_bins, l_mutau_xdsets




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import os
import os.path as op

# pip
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.ticker import (MultipleLocator)

# teslakit
from .util import MidpointNormalize
from .custom_colors import GetClusterColors

# import constants
from .config import _faspect, _fsize, _fdpi


def Generate_Covariate_Matrix(
    bmus_values, covar_values,
    bmus_dates, covar_dates,
    num_clusters, covar_rng, num_sim=1):
    'Calculates and returns matrix for stacked bar plotting'

    # generate aux arrays
    #bmus_years = [d.year for d in bmus_dates]
    m_plot = np.zeros((num_clusters, len(covar_rng)-1)) * np.nan

    for i in range(len(covar_rng)-1):

        # find years inside range
        _, s = np.where(
            [(covar_values >= covar_rng[i]) & (covar_values <= covar_rng[i+1])]
        )

        # TODO: usando alr_wrapper las fechas covar y bmus coinciden
        b = bmus_values[s,:]
        b = b.flatten()

        # TODO: mejorar, no usar los years y posicion. 
        # usar la fecha

        #ys = [covar_dates[x].year for x in s]
        # find data inside years found
        #sb = np.where(np.in1d(bmus_years, ys))[0]
        #b = bmus_values[sb,:]
        #b = b.flatten()

        for j in range(num_clusters):
            _, bb = np.where([(j+1 == b)])
            # TODO sb se utiliza para el test de laura
            #if len(sb) > 0:
                #m_plot[j,i] = float(len(bb)/float(num_sim))/len(sb)
            if len(s) > 0:
                m_plot[j,i] = float(len(bb)/float(num_sim))/len(s)
            else:
                m_plot[j,i] = 0

    return m_plot

def Generate_Covariate_rng(covar_name, cov_values):
    'Returns covar_rng and interval depending on covar_name'

    if covar_name.startswith('PC'):
        delta = 5
        n_rng = 7

        covar_rng = np.linspace(
            np.min(cov_values)-delta,
            np.max(cov_values)+delta,
            n_rng
        )

    elif covar_name.startswith('MJO'):
        delta = 0.5
        n_rng = 7

        covar_rng = np.linspace(
            np.min(cov_values)-delta,
            np.max(cov_values)+delta,
            n_rng
        )

    else:
        print('Cant plot {0}, missing rng params in plotting library'.format(
            name_covar
        ))
        return None, None

    # interval
    interval = covar_rng[1]-covar_rng[0]

    return covar_rng, interval


def Plot_PValues(p_values, term_names, show=True):
    'Plot ARL/BMUS p-values'

    n_wts = p_values.shape[0]
    n_terms = p_values.shape[1]

    # plot figure
    fig, ax = plt.subplots(1,1, figsize=(_faspect*_fsize, _fsize))

    c = ax.pcolor(p_values, cmap='coolwarm_r', clim=(0,1),
                  norm=MidpointNormalize(midpoint=0.1, vmin=0, vmax=1))
    #c.cmap.set_over('w')
    fig.colorbar(c, ax=ax)

    # Pval text
    #for i in range(p_values.shape[1]):
    #    for j in range(p_values.shape[0]):
    #        v = p_values[j,i]
    #        if v<=0.1:
    #            ax.text(i+0.5, j+0.5, '{0:.2f}'.format(v),
    #                    va='center', ha='center', size=6)

    # axis
    ax.set_title('p-value', fontweight='bold')
    ax.set_ylabel('WT')

    ax.xaxis.tick_bottom()
    ax.set_xticks(np.arange(n_terms), minor=True)
    ax.set_xticks(np.arange(n_terms)+0.5, minor=False)
    ax.set_xticklabels(term_names, minor=False, rotation=90, fontsize=7)

    ax.set_yticks(np.arange(n_wts), minor=True)
    ax.set_yticks(np.arange(n_wts)+0.5, minor=False)
    ax.set_yticklabels(np.arange(n_wts)+1, minor=False)

    # add grid
    ax.grid(True, which='minor', axis='both', linestyle='-', color='k')

    # show and return figure
    if show: plt.show()
    return fig

def Plot_Params(params, term_names, show=True):
    'Plot ARL/BMUS params'

    n_wts = params.shape[0]
    n_terms = params.shape[1]

    # plot figure
    fig, ax = plt.subplots(1,1, figsize=(_faspect*_fsize, _fsize))

    # text table and color
    #c = ax.pcolor(params, cmap=plt.cm.bwr)
    c = ax.pcolor(
        params, cmap='coolwarm_r',
        norm=MidpointNormalize(midpoint=0)
    )
    #for i in range(params.shape[1]):
    #    for j in range(params.shape[0]):
    #        v = params[j,i]
    #        ax.text(i+0.5, j+0.5, '{0:.1f}'.format(v),
    #                va='center', ha='center', size=6)
    fig.colorbar(c, ax=ax)

    # axis
    ax.set_title('params', fontweight='bold')
    ax.set_ylabel('WT')

    ax.xaxis.tick_bottom()
    ax.set_xticks(np.arange(n_terms), minor=True)
    ax.set_xticks(np.arange(n_terms)+0.5, minor=False)
    ax.set_xticklabels(term_names, minor=False, rotation=90, fontsize=7)

    ax.set_yticks(np.arange(n_wts), minor=True)
    ax.set_yticks(np.arange(n_wts)+0.5, minor=False)
    ax.set_yticklabels(np.arange(n_wts)+1, minor=False)

    # add grid
    ax.grid(True, which='minor', axis='both', linestyle='-', color='k')

    # show and return figure
    if show: plt.show()
    return fig


def Plot_Log_Sim(log, show=True):
    '''
    Plot ALR simulation log

    log - xarray.Dataset from alr wrapper (n_sim already selected)
    '''

    # plot figure
    #fig = plt.figure(figsize=(_faspect*_fsize, _fsize))
    fig = plt.figure(figsize=[18.5,9])

    # figure gridspec
    gs1 = gridspec.GridSpec(4,1)
    ax1 = fig.add_subplot(gs1[0])
    ax2 = fig.add_subplot(gs1[1], sharex=ax1)
    ax3 = fig.add_subplot(gs1[2], sharex=ax1)
    ax4 = fig.add_subplot(gs1[3], sharex=ax1)

    # Plot evbmus values
    ax1.plot(
        log.time, log.evbmus_sims, ':',
        linewidth=0.5, color='grey',
        marker='.', markersize=3,
        markerfacecolor='crimson', markeredgecolor='crimson'
    )

    ax1.yaxis.set_major_locator(MultipleLocator(4))
    ax1.grid(which='major', linestyle=':', alpha=0.5)
    ax1.set_xlim(log.time[0], log.time[-1])
    ax1.set_ylabel('Bmus', fontsize=12)

    # Plot evbmus probabilities
    z = np.diff(np.column_stack(
        ([np.zeros([len(log.time),1]), log.probTrans.values])
    ), axis=1)
    z1 = np.column_stack((z, z[:,-1])).T
    z2 = np.column_stack((z1, z1[:,-1]))
    p1 = ax2.pcolor(
        np.append(log.time, log.time[-1]),
        np.append(log.n_clusters, log.n_clusters[-1]), z2,
        cmap='PuRd', edgecolors='grey', linewidth=0.05
    )
    ax2.set_ylabel('Bmus',fontsize=12)

    # TODO: gestionar terminos markov 
    # TODO: no tengo claro si el primero oel ultimo
    alrt0 = log.alr_terms.isel(mk=0)

    # Plot Terms
    for v in range(len(log.terms)):
        if log.terms.values[v].startswith('ss'):
            ax3.plot(log.time, alrt0[:,v], label=log.terms.values[v])

        if log.terms.values[v].startswith('PC'):
            ax4.plot(log.time, alrt0[:,v], label=log.terms.values[v])

        if log.terms.values[v].startswith('MJO'):
            ax4.plot(log.time, alrt0[:,v], label=log.terms.values[v])

    # TODO: plot terms markov??

    ax3.set_ylim(-1.8,1.2)
    handles, labels = ax3.get_legend_handles_labels()
    ax3.legend(loc='lower left',ncol=len(handles))
    ax3.set_ylabel('Seasonality',fontsize=12)

    handles, labels = ax4.get_legend_handles_labels()
    ax4.legend(loc='lower left',ncol=len(handles))
    ax4.set_ylabel('Covariates',fontsize=12)
    # cbar=plt.colorbar(p1,ax=ax2,pad=0)
    # cbar.set_label('Transition probability')

    gs1.tight_layout(fig, rect=[0.05, [], 0.95, []])

    # custom colorbar for probability
    gs2 = gridspec.GridSpec(1,1)
    ax1 = fig.add_subplot(gs2[0])
    plt.colorbar(p1, cax=ax1)
    ax1.set_ylabel('Probability')
    gs2.tight_layout(fig, rect=[0.935, 0.52, 0.99, 0.73])

    # show and return figure
    if show: plt.show()
    return fig


# TODO: following functions are not finished / tested

def Plot_Covariate(bmus_values, covar_values,
                   bmus_dates, covar_dates,
                   num_clusters, name_covar,
                   num_sims=1,
                   p_export=None):
    'Plots ARL covariate related to bmus stacked bar chart'

    # get cluster colors for stacked bar plot
    np_colors_int = GetClusterColors(num_clusters)

    # generate common covar_rng
    covar_rng, interval = Generate_Covariate_rng(
        name_covar, covar_values)

    # generate plot matrix
    m_plot = Generate_Covariate_Matrix(
        bmus_values, covar_values,
        bmus_dates, covar_dates,
        num_clusters, covar_rng, num_sims)

    # plot figure
    fig, ax = plt.subplots(1,1, figsize=(_faspect*_fsize, _fsize))
    x_val = covar_rng[:-1]

    bottom_val = np.zeros(m_plot[1,:].shape)
    for r in range(num_clusters):
        row_val = m_plot[r,:]
        plt.bar(
            x_val, row_val, bottom=bottom_val,
            width=interval, color = np.array([np_colors_int[r]])
               )

        # store bottom
        bottom_val += row_val

    # axis
    plt.xlim(np.min(x_val)-interval/2, np.max(x_val)+interval/2)
    plt.ylim(0, 1)
    plt.xlabel(name_covar)
    plt.ylabel('')

    # show / export
    if not p_export:
        plt.show()

    else:
        fig.savefig(p_export, dpi=_fdpi)
        plt.close()

def Plot_Terms(terms_matrix, terms_dates, terms_names, show=True):
    'Plot terms used for ALR fitting'

    # number of terms
    n_sps = terms_matrix.shape[1]

    # custom fig size
    fsy = n_sps * 2

    # plot figure
    fig, ax_list = plt.subplots(
        n_sps, 1, sharex=True, figsize=(_faspect*_fsize, fsy)
    )

    x = terms_dates
    for i in range(n_sps):
        y = terms_matrix[:,i]
        n = terms_names[i]
        ax = ax_list[i]
        ax.plot(x, y, '.b')
        ax.set_title(n, loc='left', fontweight='bold', fontsize=10)
        ax.set_xlim(x[0], x[-1])
        ax.set_ylim(-1, 1)
        ax.grid(True, which='both')

        if n=='intercept':
            ax.set_ylim(0, 2)

    # date label
    fig.text(0.5, 0.04, 'date (y)', ha='center', fontweight='bold')
    fig.text(0.04, 0.5, 'value (-)', va='center', rotation='vertical',
             fontweight='bold')

    # show and return figure
    if show: plt.show()
    return fig

def Plot_Compare_Covariate(num_clusters,
                           bmus_values_sim, bmus_dates_sim,
                           bmus_values_hist, bmus_dates_hist,
                           cov_values_sim, cov_dates_sim,
                           cov_values_hist, cov_dates_hist,
                           name_covar,
                           n_sim = 1, p_export=None):
    'Plot simulated - historical bmus comparison related to covariate'

    # get cluster colors for stacked bar plot
    np_colors_int = GetClusterColors(num_clusters)

    # generate common covar_rng
    covar_rng, interval = Generate_Covariate_rng(
        name_covar, np.concatenate((cov_values_sim, cov_values_hist)))

    # generate plot matrix
    m_plot_sim = Generate_Covariate_Matrix(
        bmus_values_sim, cov_values_sim,
        bmus_dates_sim, cov_dates_sim,
        num_clusters, covar_rng, n_sim)

    m_plot_hist = Generate_Covariate_Matrix(
        bmus_values_hist, cov_values_hist,
        bmus_dates_hist, cov_dates_hist,
        num_clusters, covar_rng, 1)

    # plot figure
    fig, (ax_hist, ax_sim) = plt.subplots(2,1, figsize=(_faspect*_fsize, _fsize))
    x_val = covar_rng[:-1]

    # sim
    bottom_val = np.zeros(m_plot_sim[1,:].shape)
    for r in range(num_clusters):
        row_val = m_plot_sim[r,:]
        ax_sim.bar(
            x_val, row_val, bottom=bottom_val,
            width=interval, color = np.array([np_colors_int[r]])
               )

        # store bottom
        bottom_val += row_val

    # hist
    bottom_val = np.zeros(m_plot_hist[1,:].shape)
    for r in range(num_clusters):
        row_val = m_plot_hist[r,:]
        ax_hist.bar(
            x_val, row_val, bottom=bottom_val,
            width=interval, color = np_colors_int[r]
               )

        # store bottom
        bottom_val += row_val

    # axis
    # MEJORAR Y METER EL NOMBRE DE LA COVARIATE
    ax_sim.set_xlim(np.min(x_val)-interval/2, np.max(x_val)+interval/2)
    ax_sim.set_ylim(0, 1)
    ax_sim.set_title('Simulation')
    ax_sim.set_ylabel('')
    ax_sim.set_xlabel('')

    ax_hist.set_xlim(np.min(x_val)-interval/2, np.max(x_val)+interval/2)
    ax_hist.set_ylim(0, 1)
    ax_hist.set_title('Historical')
    ax_hist.set_ylabel('')
    ax_sim.set_xlabel('')

    fig.suptitle(name_covar, fontweight='bold', fontsize=12)

    # show / export
    if not p_export:
        plt.show()

    else:
        fig.savefig(p_export, dpi=_fdpi)
        plt.close()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import os.path as op
import copy
from datetime import datetime, date

# pip
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.dates as mdates
import matplotlib.colors as colors
from matplotlib import cm

# teslakit
from ..util.operations import GetBestRowsCols
from ..util.time_operations import npdt64todatetime as n2d
from ..util.time_operations import get_years_months_days
from ..kma import ClusterProbabilities
from .custom_colors import colors_dwt
from .wts import axplot_WT_Probs, axplot_WT_Hist

# import constants
from .config import _faspect, _fsize, _fdpi


def add_land_mask(ax, lon, lat, land, color):
    'addsland mask pcolormesh to existing pcolormesh'

    # select land in mask
    landc = land.copy()
    landc[np.isnan(land)]=1
    landc[land==1]=np.nan

    ax.pcolormesh(
        lon, lat, landc,
        cmap=colors.ListedColormap([color]), shading='gouraud',
    )

def axplot_EOF(ax, EOF_value, lon, lat, ttl='', land=None):
    'axes plot EOFs 2d map'

    cmap = cm.get_cmap('RdBu_r')

    # EOF pcolormesh 
    ax.pcolormesh(
        lon, lat, np.transpose(EOF_value),
        cmap=cmap, shading='gouraud',
        clim=(-1,1),
    )

    # optional mask land
    if type(land).__module__ == np.__name__:
        add_land_mask(ax, lon, lat, land, 'grey')

    # axis and title
    ax.set_title(
        ttl,
        {'fontsize': 10, 'fontweight':'bold'}
    )
    ax.tick_params(axis='both', which='major', labelsize=8)

def axplot_EOF_evolution(ax, time, EOF_evol):
    'axes plot EOFs evolution'

    # date axis locator
    yloc1 = mdates.YearLocator(1)
    yfmt = mdates.DateFormatter('%Y')

    # convert to datetime
    dtime = [n2d(t) for t in time]

    # plot EOF evolution 
    ax.plot(
        dtime, EOF_evol,
        linestyle='-', linewidth=0.5, color='black',
    )

    # configure axis
    ax.set_xlim(time[0], time[-1])
    ax.xaxis.set_major_locator(yloc1)
    ax.xaxis.set_major_formatter(yfmt)
    ax.grid(True, which='both', axis='x', linestyle='--', color='grey')
    ax.tick_params(axis='both', which='major', labelsize=8)

def axplot_DWT(ax, dwt, vmin, vmax, wt_num, land=None, wt_color=None):
    'axes plot EOFs 2d map'

    cmap = copy.deepcopy(cm.get_cmap('RdBu_r'))

    # EOF pcolormesh 
    pc = ax.pcolormesh(
        dwt,
        cmap = cmap, shading = 'gouraud',
        clim = (vmin, vmax),
    )

    # optional mask land
    if type(land).__module__ == np.__name__:
        landc = land.copy()
        landc[np.isnan(land)]=1
        landc[land==1]=np.nan
        ax.pcolormesh(
            np.flipud(landc),
            cmap=colors.ListedColormap(['silver']), shading='gouraud',
        )

    # axis color
    plt.setp(ax.spines.values(), color=wt_color)
    plt.setp(
        [ax.get_xticklines(), ax.get_yticklines()],
        color=wt_color,
    )
    for axis in ['top','bottom','left','right']:
        ax.spines[axis].set_linewidth(2.5)

    # wt text
    ax.text(0.87, 0.85, wt_num, transform=ax.transAxes, fontweight='bold')

    # customize axis
    ax.set_xticks([])
    ax.set_yticks([])

    return pc


def Plot_EOFs_EstelaPred(xds_PCA, n_plot, mask_land=None, show=True):
    '''
    Plot annual EOFs for 3D predictors

    xds_PCA:
        (n_components, n_components) PCs
        (n_components, n_features) EOFs
        (n_components, ) variance

        (n_lon, ) pred_lon: predictor longitude values
        (n_lat, ) pred_lat: predictor latitude values
        (n_time, ) pred_time: predictor time values

        method: gradient + estela

    n_plot: number of EOFs plotted
    '''

    # TODO: fix data_pos, fails only after pred.Load()?

    # PCA data
    variance = xds_PCA['variance'].values[:]
    EOFs = np.transpose(xds_PCA['EOFs'].values[:])
    PCs = np.transpose(xds_PCA['PCs'].values[:])
    data_pos = xds_PCA['pred_data_pos'].values[:]  # for handling nans
    pca_time = xds_PCA['pred_time'].values[:]
    pred_name = xds_PCA.attrs['pred_name']

    # PCA lat lon metadata
    lon = xds_PCA['pred_lon'].values
    lat = xds_PCA['pred_lat'].values

    # percentage of variance each field explains
    n_percent = variance / np.sum(variance)

    l_figs = []
    for it in range(n_plot):

        # get vargrd 
        var_grd_1d = EOFs[:,it] * np.sqrt(variance[it])

        # insert nans in data
        base = np.nan * np.ones(data_pos.shape)
        base[data_pos] = var_grd_1d

        var = base[:int(len(base)/2)]
        grd = base[int(len(base)/2):]

        # reshape data to grid
        C1 = np.reshape(var, (len(lon), len(lat)))
        C2 = np.reshape(grd, (len(lon), len(lat)))

        # figure
        fig = plt.figure(figsize=(_faspect*_fsize, 2.0/3.0*_fsize))

        # layout
        gs = gridspec.GridSpec(4, 4, wspace=0.10, hspace=0.2)

        ax_EOF_1 = plt.subplot(gs[:3, :2])
        ax_EOF_2 = plt.subplot(gs[:3, 2:])
        ax_evol = plt.subplot(gs[3, :])

        # EOF pcolormesh (SLP and GRADIENT)
        axplot_EOF(ax_EOF_1, C1, lon, lat, ttl = pred_name, land=mask_land)
        axplot_EOF(ax_EOF_2, C2, lon, lat, ttl = 'GRADIENT', land=mask_land)

        # time series EOF evolution
        evol =  PCs[it,:]/np.sqrt(variance[it])
        axplot_EOF_evolution(ax_evol, pca_time, evol)

        # figure title
        ttl = 'EOF #{0}  ---  {1:.2f}%'.format(it+1, n_percent[it]*100)
        fig.suptitle(ttl, fontsize=14, fontweight='bold')

        l_figs.append(fig)

    # show and return figure
    if show: plt.show()
    return l_figs

def Plot_ESTELA(pnt_lon, pnt_lat, estela_F, estela_D,
                lon1=None, lon2=None, lat1= None, lat2=None,
                show=True):

    'Plots ESTELA days at world map '

    try:
        from mpl_toolkits.basemap import Basemap
    except:
        print('basemap module required.')
        return

    # estela data
    estela_lon = estela_F.longitude.values[:]
    estela_lat = estela_F.latitude.values[:]
    estela_energy = estela_F.values[:]
    estela_days = estela_D.values[:]

    # energy units
    f_comp = 1
    aux_energy = np.log2(estela_energy / f_comp * 360)
    aux_energy[aux_energy<0] = 0

    # figure
    fig, ax = plt.subplots(1, figsize=(_faspect*_fsize, _fsize))

    # setup mercator map projection.
    m = Basemap(
        llcrnrlon = lon1, llcrnrlat = lat1,
        urcrnrlon = lon2, urcrnrlat = lat2,
        resolution = 'l', projection = 'cyl',
        lat_0 = pnt_lat, lon_0 = pnt_lon,
        area_thresh = 0.01,
    )
    m.drawcoastlines()
    m.fillcontinents(color = 'silver')
    m.drawmapboundary(fill_color = 'lightcyan')
    m.drawparallels(np.arange(-90, 90, 20), labels = [1,1,0,0])
    m.drawmeridians(np.arange(0, 360, 20), labels = [0,0,1,0])

    # plot estela energy (pcolormesh)
    cmap = plt.get_cmap('jet')
    pc = m.pcolormesh(
        estela_lon, estela_lat, aux_energy,
        cmap = 'jet',  shading='gouraud',
    )
    pc.set_clim(0,11)

    # colorbar
    custom_ticks = np.arange(11+1)
    cb = m.colorbar(pc, location='bottom', ticks=custom_ticks)
    cb.set_ticklabels(2**custom_ticks)
    cb.set_label(r'$\frac{kW/m}{^\circ}$ x360', x = 1.05,
                 labelpad=-30,fontsize=9)

    # plot estela days (contour)
    maxd = int(np.nanmax(estela_days))+1
    black_levels = np.arange(1, maxd, 3)
    grey_levels = np.arange(1, maxd, 1)
    grey_levels = np.setdiff1d(grey_levels, black_levels)

    ax.contour(
        estela_lon, estela_lat, estela_days,
        colors='grey', levels=grey_levels,
    )
    ac = ax.contour(
        estela_lon, estela_lat, estela_days,
        colors='k', levels=black_levels,
    )
    ax.clabel(ac, ac.levels, inline=True, fmt='%d', fontsize=10)

    # plot point
    ax.plot(pnt_lon, pnt_lat, 'ok')

    # show and return figure
    if show: plt.show()
    return fig

def Plot_DWTs_Mean_Anom(xds_KMA, xds_var, kind='mean', mask_land=None,
                        show=True):
    '''
    Plot Daily Weather Types (bmus mean)
    kind - mean/anom
    '''

    bmus = xds_KMA['sorted_bmus'].values[:]
    n_clusters = len(xds_KMA.n_clusters.values[:])

    var_max = np.max(xds_var.values)
    var_min = np.min(xds_var.values)
    scale = 1/100.0  # scale from Pa to mbar

    # get number of rows and cols for gridplot 
    n_rows, n_cols = GetBestRowsCols(n_clusters)

    # get cluster colors
    cs_dwt = colors_dwt(n_clusters)

    # plot figure
    fig = plt.figure(figsize=(_faspect*_fsize, _fsize))

    gs = gridspec.GridSpec(n_rows, n_cols, wspace=0.1, hspace=0.1)
    gr, gc = 0, 0

    for ic in range(n_clusters):

        if kind=='mean':
            # data mean
            it = np.where(bmus==ic)[0][:]
            c_mean = xds_var.isel(time=it).mean(dim='time')
            c_plot = np.multiply(c_mean, scale)  # apply scale

        elif kind=='anom':
            # data anomally
            it = np.where(bmus==ic)[0][:]
            t_mean = xds_var.mean(dim='time')
            c_mean = xds_var.isel(time=it).mean(dim='time')
            c_anom = c_mean - t_mean
            c_plot = np.multiply(c_anom, scale)  # apply scale

        # dwt color
        clr = cs_dwt[ic]

        # axes plot
        ax = plt.subplot(gs[gr, gc])
        pc = axplot_DWT(
            ax, np.flipud(c_plot),
            vmin = var_min, vmax = var_max,
            wt_num = ic+1,
            land = mask_land, wt_color = clr,
        )

        # anomalies colorbar center at 0 
        if kind == 'anom':
            pc.set_clim(-6,6)

        # get lower positions
        if gr==n_rows-1 and gc==0:
            pax_l = ax.get_position()
        elif gr==n_rows-1 and gc==n_cols-1:
            pax_r = ax.get_position()

        # counter
        gc += 1
        if gc >= n_cols:
            gc = 0
            gr += 1

    # add a colorbar        
    cbar_ax = fig.add_axes([pax_l.x0, pax_l.y0-0.05, pax_r.x1 - pax_l.x0, 0.02])
    cb = fig.colorbar(pc, cax=cbar_ax, orientation='horizontal')
    if kind=='mean':
        cb.set_label('Pressure (mbar)')
    elif kind=='anom':
        cb.set_label('Pressure anomalies (mbar)')

    # show and return figure
    if show: plt.show()
    return fig

def ClusterProbs_Month(bmus, time, wt_set, month_ix):
    'Returns Cluster probs by month_ix'

    # get months
    _, months, _ = get_years_months_days(time)

    if isinstance(month_ix, list):

        # get each month indexes
        l_ix = []
        for m_ix in month_ix:
            ixs = np.where(months == m_ix)[0]
            l_ix.append(ixs)

        # get all indexes     
        ix = np.unique(np.concatenate(tuple(l_ix)))

    else:
        ix = np.where(months == month_ix)[0]

    bmus_sel = bmus[ix]

    return ClusterProbabilities(bmus_sel, wt_set)

def Plot_DWTs_Probs(bmus, bmus_time, n_clusters, show=True):
    '''
    Plot Daily Weather Types bmus probabilities
    '''

    wt_set = np.arange(n_clusters) + 1

    # best rows cols combination
    n_rows, n_cols = GetBestRowsCols(n_clusters)

    # figure
    fig = plt.figure(figsize=(_faspect*_fsize, _fsize))

    # layout
    gs = gridspec.GridSpec(4, 7, wspace=0.10, hspace=0.25)

    # list all plots params
    l_months = [
        (1, 'January',   gs[1,3]),
        (2, 'February',  gs[2,3]),
        (3, 'March',     gs[0,4]),
        (4, 'April',     gs[1,4]),
        (5, 'May',       gs[2,4]),
        (6, 'June',      gs[0,5]),
        (7, 'July',      gs[1,5]),
        (8, 'August',    gs[2,5]),
        (9, 'September', gs[0,6]),
        (10, 'October',  gs[1,6]),
        (11, 'November', gs[2,6]),
        (12, 'December', gs[0,3]),
    ]

    l_3months = [
        ([12, 1, 2],  'DJF', gs[3,3]),
        ([3, 4, 5],   'MAM', gs[3,4]),
        ([6, 7, 8],   'JJA', gs[3,5]),
        ([9, 10, 11], 'SON', gs[3,6]),
    ]

    # plot total probabilities
    c_T = ClusterProbabilities(bmus, wt_set)
    C_T = np.reshape(c_T, (n_rows, n_cols))

    ax_probs_T = plt.subplot(gs[:2, :2])
    pc = axplot_WT_Probs(ax_probs_T, C_T, ttl = 'DWT Probabilities')

    # plot counts histogram
    ax_hist = plt.subplot(gs[2:, :3])
    axplot_WT_Hist(ax_hist, bmus, n_clusters, ttl = 'DWT Counts')

    # plot probabilities by month
    vmax = 0.15
    for m_ix, m_name, m_gs in l_months:

        # get probs matrix
        c_M = ClusterProbs_Month(bmus, bmus_time, wt_set, m_ix)
        C_M = np.reshape(c_M, (n_rows, n_cols))

        # plot axes
        ax_M = plt.subplot(m_gs)
        axplot_WT_Probs(ax_M, C_M, ttl = m_name, vmax=vmax)

    # TODO: add second colorbar?

    # plot probabilities by 3 month sets
    vmax = 0.15
    for m_ix, m_name, m_gs in l_3months:

        # get probs matrix
        c_M = ClusterProbs_Month(bmus, bmus_time, wt_set, m_ix)
        C_M = np.reshape(c_M, (n_rows, n_cols))

        # plot axes
        ax_M = plt.subplot(m_gs)
        axplot_WT_Probs(ax_M, C_M, ttl = m_name, vmax=vmax, cmap='Greens')

    # add custom colorbar
    pp = ax_probs_T.get_position()
    cbar_ax = fig.add_axes([pp.x1+0.02, pp.y0, 0.02, pp.y1 - pp.y0])
    cb = fig.colorbar(pc, cax=cbar_ax, cmap='Blues')
    cb.ax.tick_params(labelsize=8)

    # show and return figure
    if show: plt.show()
    return fig




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import time

# pip
import numpy as np
from scipy.optimize import fminbound
from sklearn.model_selection import KFold
from sklearn.metrics import mean_squared_error
import xarray as xr

# tk
from .mda import Normalize

# RBF Phi functions
def rbfphi_linear(r, const):
    return r

def rbfphi_cubic(r, const):
    return r*r*r

def rbfphi_gaussian(r, const):
    return np.exp(-0.5*r*r/(const*const))

def rbfphi_multiquadratic(r, const):
    return np.sqrt(1+r*r/(const*const))

def rbfphi_thinplate(r, const):
    return r*r*log(r+1)


def RBF_Assemble(x, phi, const, smooth):

    dim, n = x.shape
    A = np.zeros((n, n))
    for i in range(n):
        for j in range(i+1):
            r = np.linalg.norm(x[:,i]-x[:,j])
            temp = phi(r, const)
            A[i,j] = temp
            A[j,i] = temp
        A[i,i] = A[i,i] - smooth

    # polynomial part
    P = np.hstack((np.ones((n,1)), x.T))
    A = np.vstack(
        (np.hstack((A,P)), np.hstack((P.T, np.zeros((dim+1,dim+1)))))
    )

    return A

def CostEps(ep, x, y):

    # rbf coeff and A matrix calculation
    rbf_coeff, A = CalcRBF_Coeff(ep, x, y)

    # rbf cost calculation
    m, n = x.shape
    A = A[:n,:n]
    invA = np.linalg.pinv(A)
    m1, n1 = rbf_coeff.shape
    kk = y - rbf_coeff[m1-m-1]
    for i in range(m):
        kk = kk - rbf_coeff[m1-m+i] * x[i,:]
    ceps = np.dot(invA, kk) / np.diagonal(invA)
    yy = np.linalg.norm(ceps)

    return yy

def CalcRBF_Coeff(ep, x, y):

    # rbf coeff calculation
    m, n = x.shape
    A = RBF_Assemble(x, rbfphi_gaussian, ep, 0)
    b = np.concatenate((y, np.zeros((m+1,)))).reshape(-1,1)
    rbfcoeff, _, _ ,_ = np.linalg.lstsq(A, b, rcond=None)  # inverse

    return rbfcoeff, A

def RBF_Interpolation(rbf_constant, rbf_coeff, nodes, x):

    phi = rbfphi_gaussian   # gaussian RBFs
    rbf_coeff = rbf_coeff.flatten()

    dim, n = nodes.shape
    dim_p, n_p = x.shape

    f = np.zeros(n_p)
    r = np.zeros(n)

    for i in range(n_p):
        r = np.linalg.norm(
            np.repeat([x[:,i]], n, axis=0)-nodes.T,
            axis=1
        )
        s = rbf_coeff[n] + np.sum(rbf_coeff[:n] * phi(r, rbf_constant))

        # linear part
        for k in range(dim):
            s = s + rbf_coeff[k+n+1] * x[k,i]

        f[i] = s

    return f

def RBF_Reconstruction(
    subset, ix_scalar_subset, ix_directional_subset,
    target, ix_scalar_target, ix_directional_target,
    dataset):
    '''
    Radial Basis Function (Gaussian) interpolator.

    subset                - subset used for fitting RBF (dim_input)
    ix_scalar_subset      - scalar columns indexes for subset
    ix_directional_subset - directional columns indexes for subset
    target                - target used for fitting RBF (dim_output)
    ix_scalar_target      - scalar columns indexes for target
    ix_directional_target - directional columns indexes for target
    dataset - dataset used for RBF interpolation (dim_input)
    '''

    # parameters
    sigma_min = 0.1
    sigma_max = 0.7

    # normalize subset and dataset
    dataset_norm, mins, maxs = Normalize(
        dataset, ix_scalar_subset, ix_directional_subset)

    subset_norm, _, _ = Normalize(
        subset, ix_scalar_subset, ix_directional_subset, mins, maxs)

    # output storage
    output = np.zeros((dataset.shape[0], target.shape[1] ))

    # RBF scalar variables 
    for ix in ix_scalar_target:
        v = target[:,ix]

        # minimize RBF cost function
        t0 = time.time()  # time counter
        opt_sigma = fminbound(
            CostEps, sigma_min, sigma_max, args=(subset_norm.T, v)
        )
        t1 = time.time()  # optimization time

        # calculate RBF coeff
        rbf_coeff, _ = CalcRBF_Coeff(opt_sigma, subset_norm.T, v)

        # RBF interpolation
        t2 = time.time()  # time counter
        output[:, ix] = RBF_Interpolation(
            opt_sigma, rbf_coeff, subset_norm.T, dataset_norm.T)
        t3 = time.time()  # interpolation time

        print(
            'ix_scalar: {0},  optimization: {1:.2f} | interpolation: {2:.2f}'.format(
                ix, t1-t0, t3-t2)
        )

    # RBF directional variables
    for ix in ix_directional_target:
        v = target[:,ix]

        # x and y directional variable components
        vdg = np.pi/2 - v * np.pi/180
        pos = np.where(vdg < -np.pi)[0]
        vdg[pos] = vdg[pos] + 2 * np.pi
        vdx = np.cos(vdg)
        vdy = np.sin(vdg)

        # minimize RBF cost function
        t0 = time.time()  # time counter
        opt_sigma_x = fminbound(
            CostEps, sigma_min, sigma_max, args=(subset_norm.T, vdx)
        )
        opt_sigma_y = fminbound(
            CostEps, sigma_min, sigma_max, args=(subset_norm.T, vdy)
        )
        t1 = time.time()  # optimization time

        # calculate RBF coeff
        rbf_coeff_x, _ = CalcRBF_Coeff(opt_sigma_x, subset_norm.T, vdx)
        rbf_coeff_y, _ = CalcRBF_Coeff(opt_sigma_y, subset_norm.T, vdy)

        # RBF interpolation
        t2 = time.time()  # time counter
        output_x = RBF_Interpolation(
            opt_sigma_x, rbf_coeff_x, subset_norm.T, dataset_norm.T)
        output_y = RBF_Interpolation(
            opt_sigma_y, rbf_coeff_y, subset_norm.T, dataset_norm.T)
        t3 = time.time()  # interpolation time

        # join x and y components
        out = np.arctan2(output_y, output_x) * 180/np.pi
        out = 90 - out
        pos = np.where(out < 0)[0]
        out[pos] = out[pos] + 360
        output[:,ix] = out

        print(
            'ix_directional: {0},  optimization: {1:.2f} | interpolation: {2:.2f}'.format(
                ix, t1-t0, t3-t2)
        )

    return output

def RBF_Validation(
    subset, ix_scalar_subset, ix_directional_subset,
    target, ix_scalar_target, ix_directional_target,
    n_splits=3, shuffle=False):
    '''
    Radial Basis Function (Gaussian) k-fold mean squared error

    subset                - subset used for fitting RBF (dim_input)
    ix_scalar_subset      - scalar columns indexes for subset
    ix_directional_subset - directional columns indexes for subset
    target                - target used for fitting RBF (dim_output)
    ix_scalar_target      - scalar columns indexes for target
    ix_directional_target - directional columns indexes for target
    '''

    # get train-test combinations using kfold from sklearn
    kF = KFold(n_splits=n_splits, shuffle=shuffle, random_state=None)

    l_mse = []
    l_trn_ix = []
    l_tst_ix = []
    for c, (train_index, test_index) in enumerate(kF.split(subset)):
        print('RBFs Kfold Validation: {0}/{1}'.format(c+1, n_splits))

        # get train and test data
        X_train, X_test = subset[train_index], subset[test_index]
        Y_train, Y_test = target[train_index], target[test_index]

        # fit RBFs with train data and interpolate test data
        Y_rbf = RBF_Reconstruction(
            X_train, ix_scalar_subset, ix_directional_subset,
            Y_train, ix_scalar_target, ix_directional_target,
            X_test)

        # calculate mean squared error
        mse = mean_squared_error(Y_test, Y_rbf)
        print('mean squared error : {0}'.format(mse))
        print()

        # store data for output
        l_mse.append(mse)
        l_trn_ix.append(train_index)
        l_tst_ix.append(test_index)

    # return validation data
    sx = np.max([len(r) for r in l_trn_ix])
    sy = len(l_trn_ix)
    np_train = np.zeros((sx, sy)) * np.nan
    for c, r in enumerate(l_trn_ix):
        np_train[:len(r), c] = r

    sx = np.max([len(r) for r in l_tst_ix])
    sy = len(l_tst_ix)
    np_test = np.zeros((sx, sy)) * np.nan
    for c, r in enumerate(l_tst_ix):
        np_test[:len(r), c] = r

    return xr.Dataset(
        {
            'mean_squared_error': (('n_split',), l_mse),
            'train_index': (('train', 'n_split',), np_train),
            'test_index': (('test', 'n_split',), np_test),
        },
        coords={
            'n_split': np.arange(n_splits),
        }
    )




import os.path as op
import sys

import numpy as np
import pandas as pd
import xarray as xr
from wavespectra.construct.direction import cartwright
from wavespectra.construct.frequency import jonswap

from bluemath_tk.wrappers.xbeach.xbeach_wrapper import XBeachModelWrapper

example_directions = [
    7.5,
    22.5,
    37.5,
    52.5,
    67.5,
    82.5,
    97.5,
    112.5,
    127.5,
    142.5,
    157.5,
    172.5,
    187.5,
    202.5,
    217.5,
    232.5,
    247.5,
    262.5,
    277.5,
    292.5,
    307.5,
    322.5,
    337.5,
    352.5,
]
example_frequencies = [
    0.035,
    0.0385,
    0.04235,
    0.046585,
    0.0512435,
    0.05636785,
    0.06200463,
    0.0682051,
    0.07502561,
    0.08252817,
    0.090781,
    0.0998591,
    0.10984501,
    0.12082952,
    0.13291247,
    0.14620373,
    0.1608241,
    0.17690653,
    0.19459718,
    0.21405691,
    0.2354626,
    0.25900885,
    0.28490975,
    0.31340075,
    0.3447408,
    0.37921488,
    0.4171364,
    0.45885003,
    0.50473505,
]


def create_dep(filename, dep):
    with open(filename, "w") as f:
        for row in dep:
            for j in range(0, len(row), 12):
                sub_row = row[j : j + 12]
                if len(sub_row) < 12:
                    sub_row = np.append(sub_row, -999)
                for val in sub_row:
                    if val >= 0:
                        f.write(f"   {val:.7E}")
                    else:
                        f.write(f"  {val:.7E}")
                f.write("\n")


class ManuXBeachModelWrapper(XBeachModelWrapper):
    """
    Wrapper for the XBeach model.
    """

    locations = {
        "x": [
            465859.59375,
            466361.03125,
            466962.75,
            467564.46875,
        ],
        "y": [
            4810529.5,
            4809724.5,
            4808919.5,
            4808114.5,
        ],
    }

    @staticmethod
    def create_vardens(ds):
        """
        Create the vardens file from the dataset.
        """

        t = ""

        # Frequencies
        t += "{0} \n".format(len(ds.freq))
        for freq in ds.freq.values:
            t += "{0}\n".format(freq)

        # Directions
        t += "{0} \n".format(len(ds.dir))
        for dirt in sorted(ds.dir.values):
            t += "{0}\n".format(dirt)

        # Sea_surface_wave_directional_variance_spectral_density
        for _pf, freq in enumerate(ds.freq.values):
            for _pd, dirt in enumerate(sorted(ds.dir.values)):
                var = ds.sel(freq=freq, dir=dirt).efth.values
                if np.isnan(var):
                    var = 0.0
                t += "{0}\t".format(var)
            t += "\n"

        return t

    @staticmethod
    def write_file(path, text):
        with open(path, "w") as f:
            f.write(text)

    def build_case(
        self,
        case_context: dict,
        case_dir: str,
    ):
        """
        Build the XBeach case.
        """

        case_context["case_number_string"] = f"{case_context.get('case_index'):03d}"

        # /////////////////////// BATHY \\\\\\\\\\\\\\\\\\\\\\\\\
        pcs = (
            case_context.get("subset_mda")
            .isel(case=case_context.get("case_index"))[["PC1", "PC2", "PC3", "PC4"]]
            .to_array()
            .values
        )  # PCs
        anomalies = pcs @ case_context.get("eofs")  # Anomalies
        Z = np.reshape(
            anomalies + case_context.get("mean_bat").values.flatten(),
            (case_context.get("dems").x.shape[0], case_context.get("dems").x.shape[1]),
        )
        create_dep(
            op.join(
                case_dir,
                f"bathy_{case_context.get('case_index'):03d}.dep",
            ),
            Z,
        )

        # /////////////////////// SPECTRA \\\\\\\\\\\\\\\\\\\\\\\\\\
        ef = jonswap(
            freq=example_frequencies,
            fp=1
            / case_context.get("subset_mda")
            .isel(case=case_context.get("case_index"))
            .Tp.values,
            gamma=case_context.get("subset_mda")
            .isel(case=case_context.get("case_index"))
            .gamma.values,
            hs=case_context.get("subset_mda")
            .isel(case=case_context.get("case_index"))
            .Hs.values,
        )
        gth = cartwright(
            dir=example_directions,
            dm=case_context.get("subset_mda")
            .isel(case=case_context.get("case_index"))
            .Dir.values,
            dspr=case_context.get("subset_mda")
            .isel(case=case_context.get("case_index"))
            .SPR.values,
        )
        efth = ef * gth  # Offshore spectra
        off_sp = xr.Dataset(
            {"efth": (["time", "freq", "dir"], efth.expand_dims("time", axis=0).data)},
            coords={
                "dir": example_directions,
                "time": pd.to_datetime(["1996-09-23 11:00"]),
                "freq": example_frequencies,
            },
        )
        # BinWaves
        sp = off_sp.sortby(["freq", "dir"])

        for p, (x, y) in enumerate(zip(self.locations["x"], self.locations["y"])):
            kp = xr.open_dataset(
                op.join("inputs", "kp_grid", "kp_lon_{0}_lat_{1}.nc".format(x, y))
            )

            EFTH = np.full(np.shape(sp.efth.values), 0)

            for case in range(len(kp.case)):
                sys.stdout.write(
                    "Reconstructing Point {0}. Bin {1} \r".format(p + 1, case + 1)
                )
                sys.stdout.flush()

                freq_, dir_ = (
                    case_context.get("subset_mda").isel(case=case).freq.values,
                    case_context.get("subset_mda").isel(case=case).Dir.values,
                )
                efth_case = sp.sel(freq=freq_, dir=dir_, method="nearest")
                kp_case = kp.sortby("dir").isel(case=case)

                EFTH = EFTH + (efth_case.efth * kp_case.kp**2).values

            near_sp = sp.copy()
            near_sp["efth"] = (
                ("time", "freq", "dir"),
                EFTH,
            )  # Nearshore spectra (reconstructed)

            # Save vardens
            vardens = near_sp.copy()
            vardens["dir"] = (270 - (vardens["dir"])) % 360
            spec = self.create_vardens(vardens.isel(time=0))
            self.write_file(
                op.join(case_dir, "vardens{0}.txt").format(p + 1),
                spec,
            )



from . import alr
from . import awt
from . import buoy
from . import climate_change
from . import config
from . import csiro
from . import custom_colors
from . import estela
from . import extremes
from . import intradaily
from . import mda
from . import mjo
from . import outputs
from . import pcs
from . import ss
from . import storms
from . import tides
from . import util
from . import waves
from . import wts

